from pathlib import Path

from fpdf import FPDF

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_wraps_zerowidthspace(tmp_path):
    pdf = FPDF()
    pdf.add_font(fname=HERE / "Waree.ttf")
    pdf.set_font("Waree", size=12)
    pdf.add_page()
    pdf.write(
        8,
        "Thai (ideally wouldn't wrap after the space after '1000'): "
        + "นโยบาย\u200bสาธารณะ\u200bมี\u200bความ\u200bสำคัญ\u200bต่อ\u200b"
        + "การ\u200bสนับสนุน\u200bการ\u200bออก\u200bแบบ\u200bและ\u200bการ"
        + "\u200bสร้าง\u200bชุมชน\u200bและ\u200bเมือง\u200bสุขภาพ\u200bดี\u200b"
        + "และ\u200bยั่งยืน รายการ\u200bตรวจ\u200bสอบนโยบาย\u200bความ\u200b"
        + "ท้าทาย 1,000 เมือง\u200bสำหรับ\u200bใช้\u200bเพื่อ\u200bประเมิน\u200bการ"
        + "\u200bมี\u200bอยู่\u200bและ\u200bคุณภาพ\u200bของ\u200bนโยบาย\u200bที่"
        + "\u200bสอด\u200bคล้อง\u200bกับ\u200bหลัก\u200bฐาน\u200bและ\u200bหลัก"
        + "\u200bการ\u200bสำหรับ\u200bเมือง\u200bที่\u200bมี\u200bสุขภาพ\u200bดี"
        + "\u200bและ\u200bยั่งยืน",
    )
    assert_pdf_equal(pdf, HERE / "thai_wraps_zerowidthspace.pdf", tmp_path)



from pathlib import Path

import pytest
from fpdf import FPDF, FPDFException
from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"
PNG_DIR = HERE.parent / "image/png_images"
SVG_DIR = HERE.parent / "svg/svg_sources"


def test_tcols_align(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    col = pdf.text_columns()
    with col:
        col.write(text=LOREM_IPSUM[:100])
        pdf.set_font("Times", size=12)
        col.write(text=LOREM_IPSUM[100:200])
        pdf.set_font("Courier", size=12)
        col.write(text=LOREM_IPSUM[200:300])
    pdf.set_font("Helvetica", style="I", size=12)
    with col:
        with col.paragraph(text_align="J", top_margin=pdf.font_size * 2) as par:
            par.write(text=LOREM_IPSUM[:100])
            pdf.set_font("Times", style="I", size=12)
            par.write(text=LOREM_IPSUM[100:200])
            pdf.set_font("Courier", style="I", size=12)
            par.write(text=LOREM_IPSUM[200:300])
    pdf.set_font("Helvetica", style="B", size=12)
    with col:
        with col.paragraph(text_align="R", top_margin=pdf.font_size * 2) as par:
            par.write(text=LOREM_IPSUM[:100])
            pdf.set_font("Times", style="B", size=12)
            par.write(text=LOREM_IPSUM[100:200])
            pdf.set_font("Courier", style="B", size=12)
            par.write(text=LOREM_IPSUM[200:300])
    pdf.set_font("Helvetica", style="BI", size=12)
    with col:
        with col.paragraph(text_align="C", top_margin=pdf.font_size * 2) as par:
            par.write(text=LOREM_IPSUM[:100])
            pdf.set_font("Times", style="BI", size=12)
            par.write(text=LOREM_IPSUM[100:200])
            pdf.set_font("Courier", style="BI", size=12)
            par.write(text=LOREM_IPSUM[200:300])
    assert_pdf_equal(pdf, HERE / "tcols_align.pdf", tmp_path)


def test_tcols_3cols(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.t_margin = 50
    pdf.set_auto_page_break(True, 100)
    pdf.set_font("Helvetica", size=6)
    cols = pdf.text_columns(
        text=LOREM_IPSUM,
        text_align="J",
        img=SVG_DIR / "SVG_logo.svg",
        img_fill_width=True,
        ncols=3,
        gutter=5,
    )
    with cols:
        pdf.set_font("Times", size=8)
        cols.write(text=LOREM_IPSUM)
        pdf.set_font("Courier", size=10)
        cols.write(text=LOREM_IPSUM)
        pdf.set_font("Helvetica", size=12)
        cols.write(text=LOREM_IPSUM)
        pdf.set_font("Times", size=14)
        cols.write(text=LOREM_IPSUM)
        pdf.set_font("Courier", size=16)
        cols.write(text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "tcols_3cols.pdf", tmp_path)


def test_tcols_balance(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_auto_page_break(True, 100)
    pdf.set_font("Helvetica", size=6)
    cols_2 = pdf.text_columns(text_align="J", ncols=2, gutter=10, balance=True)
    cols_3 = pdf.text_columns(text_align="J", ncols=3, gutter=5, balance=True)
    with cols_2:
        pdf.set_font("Times", size=8)
        cols_2.write(text=LOREM_IPSUM[:300])
    with cols_3:
        pdf.set_font("Courier", size=10)
        cols_3.write(text=LOREM_IPSUM[300:600])
    with cols_2:
        pdf.set_font("Helvetica", size=12)
        cols_2.write(text=LOREM_IPSUM[600:900])
    with cols_3:
        pdf.set_font("Times", size=14)
        cols_3.write(text=LOREM_IPSUM[:300])
    assert_pdf_equal(pdf, HERE / "tcols_balance.pdf", tmp_path)


def test_tcols_charwrap(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("courier", size=16)
    col = pdf.text_columns(l_margin=50, r_margin=50)
    # wrapmode on paragraph
    with col.paragraph(wrapmode="CHAR", bottom_margin=pdf.font_size) as par:
        par.write(text=LOREM_IPSUM[:500])
    col.render()
    # wrapmode on column
    with pdf.text_columns(
        l_margin=50,
        r_margin=50,
        wrapmode="CHAR",
    ) as col:
        with col.paragraph() as par:
            par.write(text=LOREM_IPSUM[500:1000])
    assert_pdf_equal(pdf, HERE / "tcols_charwrap.pdf", tmp_path)


def test_tcols_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=12)
    cols = pdf.text_columns(ncols=3, text_align="J")
    left, right = cols.current_x_extents(pdf.y, 0)
    pdf.line(left, pdf.t_margin, left, pdf.h - pdf.b_margin)
    pdf.line(right, pdf.t_margin, right, pdf.h - pdf.b_margin)
    with cols:
        cols.write(text="Images at Native Size\n(Raster 72 dpi, SVG 90 dpi)\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png")
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(SVG_DIR / "SVG_logo.svg")
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Images at Full Column Width\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            fill_width=True,
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(SVG_DIR / "SVG_logo.svg", fill_width=True)
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Images Aligned Right and Center\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            align="RIGHT",
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(SVG_DIR / "SVG_logo.svg", align="CENTER")
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Images at Size 50x20\ntop/bottom margin\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            width=50,
            height=20,
            top_margin=10,
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(
            SVG_DIR / "SVG_logo.svg",
            width=50,
            height=20,
            bottom_margin=10,
        )
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Images at Size 50x20 and keep_aspect_ratio=True\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            width=50,
            height=20,
            keep_aspect_ratio=True,
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(
            SVG_DIR / "SVG_logo.svg",
            width=50,
            height=20,
            keep_aspect_ratio=True,
        )
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Images at Size 20x50 and keep_aspect_ratio=True\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            width=20,
            height=50,
            keep_aspect_ratio=True,
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(
            SVG_DIR / "SVG_logo.svg",
            width=20,
            height=50,
            keep_aspect_ratio=True,
        )
        cols.write(text=LOREM_IPSUM[200:300])

        cols.new_column()
        cols.write(text="Column break by image\n\n")
        cols.write(text=LOREM_IPSUM[:100])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            fill_width=True,
        )
        cols.write(text=LOREM_IPSUM[100:200])
        cols.image(SVG_DIR / "SVG_logo.svg", fill_width=True)
        cols.write(text=LOREM_IPSUM[200:300])
        cols.image(
            PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png",
            fill_width=True,
        )
        cols.image(SVG_DIR / "SVG_logo.svg", fill_width=True)

    assert_pdf_equal(pdf, HERE / "tcols_images.pdf", tmp_path)


def test_tcols_no_font():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        col = pdf.text_columns()
        col.write("something")
    expected_msg = "No font set, you need to call set_font() beforehand"
    assert str(error.value) == expected_msg
    with pytest.raises(FPDFException) as error:
        col.ln()
    expected_msg = "No font set, you need to call set_font() beforehand"
    assert str(error.value) == expected_msg


def test_tcols_bad_uses():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("courier", size=16)
    col = pdf.text_columns()
    # recursive text region context
    with col:
        col.write("something")
        with pytest.raises(FPDFException) as error:
            with col:
                pass
    expected_msg = "Unable to enter the same TextColumns context recursively."
    assert str(error.value) == expected_msg
    # recursive use of paragraph context
    with col.paragraph() as par:
        par.write("something")
        with pytest.raises(FPDFException) as error:
            col.paragraph()
    expected_msg = "Unable to nest paragraphs."
    assert str(error.value) == expected_msg
    # writing to column while we have an explicit paragraph active
    with col.paragraph() as par:
        par.write("something")
        with pytest.raises(FPDFException) as error:
            col.write("else")
    expected_msg = "Conflicts with active paragraph. Either close the current paragraph or write your text inside it."
    assert str(error.value) == expected_msg
    # ending a non-existent paragraph
    with pytest.raises(FPDFException) as error:
        col.end_paragraph()
    expected_msg = "No active paragraph to end."
    assert str(error.value) == expected_msg
    # column with negative width
    with pytest.raises(FPDFException) as error:
        col = pdf.text_columns(l_margin=150, r_margin=150)
    expected_msg = (
        "TextColumns(): Right limit (60.00155555555551) lower than left limit (150)."
    )
    assert str(error.value) == expected_msg
    # invalid alignment values
    with pytest.raises(ValueError) as error:
        col = pdf.text_columns(text_align="X_CENTER")
    expected_msg = (
        "Text_align must be 'LEFT', 'CENTER', 'RIGHT', or 'JUSTIFY', not 'X_CENTER'."
    )
    assert str(error.value) == expected_msg
    with pytest.raises(ValueError) as error:
        with pdf.text_columns() as col:
            par = col.paragraph(text_align="X_CENTER")
    assert str(error.value) == expected_msg
    with pytest.raises(ValueError) as error:
        with pdf.text_columns() as col:
            col.image(name="foo", align="JUSTIFY")
    expected_msg = "Align must be 'LEFT', 'CENTER', or 'RIGHT', not 'JUSTIFY'."
    assert str(error.value) == expected_msg
    # Wrong call sequence for ImageParagraph.
    with pdf.text_columns() as col:
        col.image(name=PNG_DIR / "ac6343a98f8edabfcc6e536dd75aacb0.png")
        with pytest.raises(RuntimeError) as error:
            col._paragraphs[-1].render(1, 2, 3)  # pylint: disable=protected-access
    expected_msg = "ImageParagraph.build_line() must be called before render()."
    assert str(error.value) == expected_msg


def test_tcols_break_top_margin(tmp_path):  # regression test for #1214
    """Ensure that the top/bottom margins of a paragraph trigger
    a column break.
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=14)
    pdf.set_top_margin(50)
    pdf.set_auto_page_break(True, 50)
    pdf.rect(pdf.l_margin, pdf.t_margin, pdf.epw, pdf.eph)

    with pdf.text_columns(text_align="J", ncols=2) as cols:
        for _ in range(3):
            with cols.paragraph(
                text_align="J",
                bottom_margin=pdf.font_size * 5,
                top_margin=pdf.font_size * 5,
            ) as par:
                par.write(text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "tcols_break_top_margin.pdf", tmp_path)


def test_paragraph_emphasis(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=32)
    column = pdf.text_columns()
    # Paragraph 1:
    par = column.paragraph()
    par.write(text="Un.")
    column.end_paragraph()
    # Paragraph 2:
    pdf.set_font(style="B")
    par = column.paragraph()
    par.write(text="Deux.")
    pdf.set_font(style="I")
    par.write(text="Trois.")
    column.end_paragraph()
    column.render()
    assert_pdf_equal(pdf, HERE / "paragraph_emphasis.pdf", tmp_path)


def test_paragraph_first_line_indent(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=12)
    with pdf.text_columns() as cols:
        with cols.paragraph(first_line_indent=10, text_align="J") as paragraph:
            paragraph.write(text=LOREM_IPSUM)
    assert_pdf_equal(pdf, HERE / "paragraph_first_line_indent.pdf", tmp_path)


def test_text_columns_with_text_shaping(tmp_path):  # issue 1439
    pdf = FPDF()
    pdf.add_page()
    pdf.set_text_shaping(True)
    pdf.set_font("Times", "", 14)
    with pdf.text_columns(ncols=2) as cols:
        cols.write("Lorem ipsum dolor sit amet")
        cols.new_column()
        cols.write("Lorem ipsum dolor sit amet")
    assert_pdf_equal(pdf, HERE / "text_columns_with_text_shaping.pdf", tmp_path)


def test_text_columns_with_shorter_2nd_column(tmp_path):  # issue 1442
    pdf = FPDF()
    pdf.add_page()
    pdf.set_text_shaping(True)
    pdf.set_font("Times", size=14)
    pdf.write(text="Headline")
    pdf.ln()
    pdf.ln()
    with pdf.text_columns(ncols=2) as cols:
        cols.write(
            "Lorem ipsum dolor sit amet, consetetur sadipscing elitr,"
            " sed diam nonumy eirmod tempor invidunt ut labore"
            " et dolore magna aliquyam erat, sed diam voluptua."
        )
        cols.new_column()
        cols.write("Lorem ipsum dolor sit amet")
    pdf.ln()
    pdf.write(text="More text after columns.")
    pdf.ln()
    assert_pdf_equal(pdf, HERE / "text_columns_with_shorter_2nd_column.pdf", tmp_path)



# This is an implementation of the Unicode Standard Annex #9
# Unicode bidirectional algorithm - Revision 48 for Unicode 15.1.0
# https://unicode.org/reports/tr9/

import unicodedata
from collections import deque
from dataclasses import dataclass, replace
from operator import itemgetter
from typing import Optional, TypedDict

from .enums import TextDirection

MAX_DEPTH: int = 125

# BidiBrackets 15.1.0 2023-01-18
# Loaded from https://www.unicode.org/Public/UNIDATA/BidiBrackets.txt
# This table can be dropped when the information is added on "unicodedata"
BIDI_BRACKETS: dict[str, dict[str, str]] = {
    "(": {"pair": ")", "type": "o"},
    ")": {"pair": "(", "type": "c"},
    "[": {"pair": "]", "type": "o"},
    "]": {"pair": "[", "type": "c"},
    "{": {"pair": "}", "type": "o"},
    "}": {"pair": "{", "type": "c"},
    "༺": {"pair": "༻", "type": "o"},
    "༻": {"pair": "༺", "type": "c"},
    "༼": {"pair": "༽", "type": "o"},
    "༽": {"pair": "༼", "type": "c"},
    "᚛": {"pair": "᚜", "type": "o"},
    "᚜": {"pair": "᚛", "type": "c"},
    "⁅": {"pair": "⁆", "type": "o"},
    "⁆": {"pair": "⁅", "type": "c"},
    "⁽": {"pair": "⁾", "type": "o"},
    "⁾": {"pair": "⁽", "type": "c"},
    "₍": {"pair": "₎", "type": "o"},
    "₎": {"pair": "₍", "type": "c"},
    "⌈": {"pair": "⌉", "type": "o"},
    "⌉": {"pair": "⌈", "type": "c"},
    "⌊": {"pair": "⌋", "type": "o"},
    "⌋": {"pair": "⌊", "type": "c"},
    "〈": {"pair": "〉", "type": "o"},
    "〉": {"pair": "〈", "type": "c"},
    "❨": {"pair": "❩", "type": "o"},
    "❩": {"pair": "❨", "type": "c"},
    "❪": {"pair": "❫", "type": "o"},
    "❫": {"pair": "❪", "type": "c"},
    "❬": {"pair": "❭", "type": "o"},
    "❭": {"pair": "❬", "type": "c"},
    "❮": {"pair": "❯", "type": "o"},
    "❯": {"pair": "❮", "type": "c"},
    "❰": {"pair": "❱", "type": "o"},
    "❱": {"pair": "❰", "type": "c"},
    "❲": {"pair": "❳", "type": "o"},
    "❳": {"pair": "❲", "type": "c"},
    "❴": {"pair": "❵", "type": "o"},
    "❵": {"pair": "❴", "type": "c"},
    "⟅": {"pair": "⟆", "type": "o"},
    "⟆": {"pair": "⟅", "type": "c"},
    "⟦": {"pair": "⟧", "type": "o"},
    "⟧": {"pair": "⟦", "type": "c"},
    "⟨": {"pair": "⟩", "type": "o"},
    "⟩": {"pair": "⟨", "type": "c"},
    "⟪": {"pair": "⟫", "type": "o"},
    "⟫": {"pair": "⟪", "type": "c"},
    "⟬": {"pair": "⟭", "type": "o"},
    "⟭": {"pair": "⟬", "type": "c"},
    "⟮": {"pair": "⟯", "type": "o"},
    "⟯": {"pair": "⟮", "type": "c"},
    "⦃": {"pair": "⦄", "type": "o"},
    "⦄": {"pair": "⦃", "type": "c"},
    "⦅": {"pair": "⦆", "type": "o"},
    "⦆": {"pair": "⦅", "type": "c"},
    "⦇": {"pair": "⦈", "type": "o"},
    "⦈": {"pair": "⦇", "type": "c"},
    "⦉": {"pair": "⦊", "type": "o"},
    "⦊": {"pair": "⦉", "type": "c"},
    "⦋": {"pair": "⦌", "type": "o"},
    "⦌": {"pair": "⦋", "type": "c"},
    "⦍": {"pair": "⦐", "type": "o"},
    "⦎": {"pair": "⦏", "type": "c"},
    "⦏": {"pair": "⦎", "type": "o"},
    "⦐": {"pair": "⦍", "type": "c"},
    "⦑": {"pair": "⦒", "type": "o"},
    "⦒": {"pair": "⦑", "type": "c"},
    "⦓": {"pair": "⦔", "type": "o"},
    "⦔": {"pair": "⦓", "type": "c"},
    "⦕": {"pair": "⦖", "type": "o"},
    "⦖": {"pair": "⦕", "type": "c"},
    "⦗": {"pair": "⦘", "type": "o"},
    "⦘": {"pair": "⦗", "type": "c"},
    "⧘": {"pair": "⧙", "type": "o"},
    "⧙": {"pair": "⧘", "type": "c"},
    "⧚": {"pair": "⧛", "type": "o"},
    "⧛": {"pair": "⧚", "type": "c"},
    "⧼": {"pair": "⧽", "type": "o"},
    "⧽": {"pair": "⧼", "type": "c"},
    "⸢": {"pair": "⸣", "type": "o"},
    "⸣": {"pair": "⸢", "type": "c"},
    "⸤": {"pair": "⸥", "type": "o"},
    "⸥": {"pair": "⸤", "type": "c"},
    "⸦": {"pair": "⸧", "type": "o"},
    "⸧": {"pair": "⸦", "type": "c"},
    "⸨": {"pair": "⸩", "type": "o"},
    "⸩": {"pair": "⸨", "type": "c"},
    "⹕": {"pair": "⹖", "type": "o"},
    "⹖": {"pair": "⹕", "type": "c"},
    "⹗": {"pair": "⹘", "type": "o"},
    "⹘": {"pair": "⹗", "type": "c"},
    "⹙": {"pair": "⹚", "type": "o"},
    "⹚": {"pair": "⹙", "type": "c"},
    "⹛": {"pair": "⹜", "type": "o"},
    "⹜": {"pair": "⹛", "type": "c"},
    "〈": {"pair": "〉", "type": "o"},
    "〉": {"pair": "〈", "type": "c"},
    "《": {"pair": "》", "type": "o"},
    "》": {"pair": "《", "type": "c"},
    "「": {"pair": "」", "type": "o"},
    "」": {"pair": "「", "type": "c"},
    "『": {"pair": "』", "type": "o"},
    "』": {"pair": "『", "type": "c"},
    "【": {"pair": "】", "type": "o"},
    "】": {"pair": "【", "type": "c"},
    "〔": {"pair": "〕", "type": "o"},
    "〕": {"pair": "〔", "type": "c"},
    "〖": {"pair": "〗", "type": "o"},
    "〗": {"pair": "〖", "type": "c"},
    "〘": {"pair": "〙", "type": "o"},
    "〙": {"pair": "〘", "type": "c"},
    "〚": {"pair": "〛", "type": "o"},
    "〛": {"pair": "〚", "type": "c"},
    "﹙": {"pair": "﹚", "type": "o"},
    "﹚": {"pair": "﹙", "type": "c"},
    "﹛": {"pair": "﹜", "type": "o"},
    "﹜": {"pair": "﹛", "type": "c"},
    "﹝": {"pair": "﹞", "type": "o"},
    "﹞": {"pair": "﹝", "type": "c"},
    "（": {"pair": "）", "type": "o"},
    "）": {"pair": "（", "type": "c"},
    "［": {"pair": "］", "type": "o"},
    "］": {"pair": "［", "type": "c"},
    "｛": {"pair": "｝", "type": "o"},
    "｝": {"pair": "｛", "type": "c"},
    "｟": {"pair": "｠", "type": "o"},
    "｠": {"pair": "｟", "type": "c"},
    "｢": {"pair": "｣", "type": "o"},
    "｣": {"pair": "｢", "type": "c"},
}


class BidiCharacter:
    __slots__ = [
        "character_index",
        "character",
        "bidi_class",
        "original_bidi_class",
        "embedding_level",
        "direction",
    ]

    def __init__(
        self, character_index: int, character: str, embedding_level: int, debug: bool
    ) -> None:
        self.character_index = character_index
        self.character = character
        if debug and character.isupper():
            self.bidi_class = "R"
        else:
            self.bidi_class = unicodedata.bidirectional(character)
        self.original_bidi_class = self.bidi_class
        self.embedding_level = embedding_level
        self.direction = None

    def get_direction_from_level(self) -> str:
        return "R" if self.embedding_level % 2 else "L"

    def set_class(self, cls: str) -> None:
        self.bidi_class = cls

    def __repr__(self) -> str:
        return (
            f"character_index: {self.character_index} character: {self.character}"
            + f" bidi_class: {self.bidi_class} original_bidi_class: {self.original_bidi_class}"
            + f" embedding_level: {self.embedding_level} direction: {self.direction}"
        )


@dataclass
class DirectionalStatus:
    __slots__ = [
        "embedding_level",
        "directional_override_status",
        "directional_isolate_status",
    ]
    embedding_level: int  # between 0 and MAX_DEPTH
    directional_override_status: str  # "N" (Neutral), "L" (Left) or "R" (Right)
    directional_isolate_status: bool


class IsolatingRun:
    __slots__ = ["characters", "previous_direction", "next_direction"]

    def __init__(self, characters: list[BidiCharacter], sos: str, eos: str) -> None:
        self.characters = characters
        self.previous_direction = sos
        self.next_direction = eos
        self.resolve_weak_types()
        self.resolve_neutral_types()
        self.resolve_implicit_levels()

    def resolve_weak_types(self) -> None:
        # W1. Examine each nonspacing mark (NSM) in the isolating run sequence, and change the type of the NSM to Other Neutral
        #     if the previous character is an isolate initiator or PDI, and to the type of the previous character otherwise.
        #     If the NSM is at the start of the isolating run sequence, it will get the type of sos.
        for i, bidi_char in enumerate(self.characters):
            if bidi_char.bidi_class == "NSM":
                if i == 0:
                    bidi_char.set_class(self.previous_direction)
                else:
                    bidi_char.set_class(
                        "ON"
                        if self.characters[i - 1].bidi_class
                        in ("LRI", "RLI", "FSI", "PDI")
                        else self.characters[i - 1].bidi_class
                    )

        # W2. Search backward from each instance of a European number until the first strong type (R, L, AL, or sos) is found.
        #     If an AL is found, change the type of the European number to Arabic number.
        # W3. Change all ALs to R.

        last_strong_type = self.previous_direction
        for bidi_char in self.characters:
            if bidi_char.bidi_class in ("R", "L", "AL"):
                last_strong_type = bidi_char.bidi_class
            if bidi_char.bidi_class == "AL":
                bidi_char.set_class("R")
            if bidi_char.bidi_class == "EN" and last_strong_type == "AL":
                bidi_char.set_class("AN")

        # W4. A single European separator between two European numbers changes to a European number.
        #     A single common separator between two numbers of the same type changes to that type.
        for i, bidi_char in enumerate(self.characters):
            if i in (0, len(self.characters) - 1):
                continue
            if (
                bidi_char.bidi_class == "ES"
                and self.characters[i - 1].bidi_class == "EN"
                and self.characters[i + 1].bidi_class == "EN"
            ):
                bidi_char.set_class("EN")

            if (
                bidi_char.bidi_class == "CS"
                and self.characters[i - 1].bidi_class in ("AN", "EN")
                and self.characters[i + 1].bidi_class
                == self.characters[i - 1].bidi_class
            ):
                bidi_char.set_class(self.characters[i - 1].bidi_class)

        # W5. A sequence of European terminators adjacent to European numbers changes to all European numbers.
        # W6. All remaining separators and terminators (after the application of W4 and W5) change to Other Neutral.
        def prev_is_en(i: int) -> bool:
            if i == 0:
                return False
            if self.characters[i - 1].bidi_class == "ET":
                return prev_is_en(i - 1)
            return self.characters[i - 1].bidi_class == "EN"

        def next_is_en(i: int) -> bool:
            if i == len(self.characters) - 1:
                return False
            if self.characters[i + 1].bidi_class == "ET":
                return next_is_en(i + 1)
            return self.characters[i + 1].bidi_class == "EN"

        for i, bidi_char in enumerate(self.characters):
            if bidi_char.bidi_class == "ET":
                if prev_is_en(i) or next_is_en(i):
                    bidi_char.set_class("EN")

            if bidi_char.bidi_class in ("ET", "ES", "CS"):
                bidi_char.set_class("ON")
        # W7. Search backward from each instance of a European number until the first strong type (R, L, or sos) is found.
        #     If an L is found, then change the type of the European number to L.
        last_strong_type = self.previous_direction
        for bidi_char in self.characters:
            if bidi_char.bidi_class in ("R", "L", "AL"):
                last_strong_type = bidi_char.bidi_class
            if bidi_char.bidi_class == "EN" and last_strong_type == "L":
                bidi_char.set_class("L")

    def pair_brackets(self) -> list[tuple[int, int]]:
        """
        Calculate all the bracket pairs on an isolate run, to be used on rule N0
        How to calculate bracket pairs:
        - Basic definitions 14, 15 and 16: http://www.unicode.org/reports/tr9/#BD14
        - BIDI brackets for dummies: https://www.unicode.org/notes/tn39/
        """
        open_brackets: list[tuple[str, int]] = []
        open_bracket_count = 0
        bracket_pairs: list[tuple[int, int]] = []
        for index, char in enumerate(self.characters):
            if char.character in BIDI_BRACKETS and char.bidi_class == "ON":
                if BIDI_BRACKETS[char.character]["type"] == "o":
                    if open_bracket_count >= 63:
                        return []
                    open_brackets.append((char.character, index))
                    open_bracket_count += 1
                if BIDI_BRACKETS[char.character]["type"] == "c":
                    if open_bracket_count == 0:
                        continue
                    for current_open_bracket in range(open_bracket_count, 0, -1):
                        open_char, open_index = open_brackets[current_open_bracket - 1]
                        if (BIDI_BRACKETS[open_char]["pair"] == char.character) or (
                            BIDI_BRACKETS[open_char]["pair"] in ("〉", "〉")
                            and char.character in ("〉", "〉")
                        ):
                            bracket_pairs.append((open_index, index))
                            open_brackets = open_brackets[: current_open_bracket - 1]
                            open_bracket_count = current_open_bracket - 1
                            break
        return sorted(bracket_pairs, key=itemgetter(0))

    def resolve_neutral_types(self) -> None:
        def previous_strong(index: int) -> str:
            if index == 0:
                return self.previous_direction
            if self.characters[index - 1].bidi_class == "L":
                return "L"
            if self.characters[index - 1].bidi_class in ("R", "AN", "EN"):
                return "R"
            return previous_strong(index - 1)

        def next_strong(index: int) -> str:
            if index >= len(self.characters) - 1:
                return self.next_direction
            if self.characters[index + 1].bidi_class == "L":
                return "L"
            if self.characters[index + 1].bidi_class in ("R", "AN", "EN"):
                return "R"
            return next_strong(index + 1)

        # N0-N2: Resolving neutral types
        # N0
        brackets = self.pair_brackets()
        if brackets:
            embedding_direction = self.characters[0].get_direction_from_level()
            for b in brackets:
                strong_same_direction = False
                strong_opposite_direction = False
                resulting_direction = None
                for index in range(b[0], b[1]):
                    if (
                        self.characters[index].bidi_class == "L"
                        and embedding_direction == "L"
                    ) or (
                        self.characters[index].bidi_class in ("R", "AN", "EN")
                        and embedding_direction == "R"
                    ):
                        strong_same_direction = True
                        break
                    if (
                        self.characters[index].bidi_class == "L"
                        and embedding_direction == "R"
                    ) or (
                        self.characters[index].bidi_class in ("R", "AN", "EN")
                        and embedding_direction == "L"
                    ):
                        strong_opposite_direction = True
                if strong_same_direction:
                    resulting_direction = embedding_direction
                elif strong_opposite_direction:
                    opposite_direction = "L" if embedding_direction == "R" else "R"
                    if previous_strong(b[0]) == opposite_direction:
                        resulting_direction = opposite_direction
                    else:
                        resulting_direction = embedding_direction
                if resulting_direction:
                    self.characters[b[0]].bidi_class = resulting_direction
                    self.characters[b[1]].bidi_class = resulting_direction
                    if len(self.characters) > b[1] + 1:
                        next_char = self.characters[b[1] + 1]
                        if (
                            next_char.original_bidi_class == "NSM"
                            and next_char.bidi_class == "ON"
                        ):
                            next_char.bidi_class = resulting_direction

        for i, bidi_char in enumerate(self.characters):
            # N1-N2
            if bidi_char.bidi_class in (
                "B",
                "S",
                "WS",
                "ON",
                "FSI",
                "LRI",
                "RLI",
                "PDI",
            ):
                if previous_strong(i) == next_strong(i):
                    bidi_char.bidi_class = previous_strong(i)
                else:
                    bidi_char.bidi_class = bidi_char.get_direction_from_level()

    def resolve_implicit_levels(self) -> None:
        for bidi_char in self.characters:
            # I1. For all characters with an even (left-to-right) embedding level,
            #     those of type R go up one level and those of type AN or EN go up two levels.
            if bidi_char.embedding_level % 2 == 0:
                if bidi_char.bidi_class == "R":
                    bidi_char.embedding_level += 1
                if bidi_char.bidi_class in ("AN", "EN"):
                    bidi_char.embedding_level += 2

            # I2. For all characters with an odd (right-to-left) embedding level, those of type L, EN or AN go up one level.
            else:
                if bidi_char.bidi_class in ("L", "EN", "AN"):
                    bidi_char.embedding_level += 1


def auto_detect_base_direction(
    string: str, stop_at_pdi: bool = False, debug: bool = False
) -> TextDirection:
    """
    This function applies rules P2 and P3 to detect the direction of a paragraph, retuning
    the first strong direction and skipping over isolate sequences.
    P1 must be applied before calling this function (breaking into paragraphs)
    stop_at_pdi can be set to True to get the direction of a single isolate sequence
    """
    # Auto-LTR (standard BIDI) uses the first L/R/AL character, and is LTR if none is found.
    isolate = 0
    for char in string:
        bidi_class = unicodedata.bidirectional(char)
        if debug and bidi_class.isupper():
            bidi_class = "R"
        if bidi_class == "PDI" and isolate == 0 and stop_at_pdi:
            return TextDirection.LTR
        if bidi_class in ("LRI", "RLI", "FSI"):
            isolate += 1
        if bidi_class == "PDI" and isolate > 0:
            isolate -= 1
        if bidi_class in ("R", "AL") and isolate == 0:
            return TextDirection.RTL
        if bidi_class == "L" and isolate == 0:
            return TextDirection.LTR
    return TextDirection.LTR


def calculate_isolate_runs(paragraph: list[BidiCharacter]) -> list[IsolatingRun]:
    # BD13 and X10
    class LevelRun(TypedDict):
        level: int
        text: list[BidiCharacter]
        complete: bool
        sos: str
        eos: str

    level_run: list[LevelRun] = []
    lr: list[BidiCharacter] = []
    lr_embedding_level: int = paragraph[0].embedding_level

    for bidi_char in paragraph:
        if bidi_char.embedding_level != lr_embedding_level:
            level_run.append(
                {
                    "level": lr_embedding_level,
                    "text": lr,
                    "complete": False,
                    "sos": "",
                    "eos": "",
                }
            )
            lr = []
            lr_embedding_level = bidi_char.embedding_level
        lr.append(bidi_char)
    level_run.append(
        {
            "level": lr_embedding_level,
            "text": lr,
            "complete": False,
            "sos": "",
            "eos": "",
        }
    )

    def level_to_direction(level: int) -> str:
        if level % 2 == 0:
            return "L"
        return "R"

    # compute sos, eos for each level run
    for index, lr1 in enumerate(level_run):
        if lr1["complete"]:
            continue
        if index == 0:
            sos = level_to_direction(lr1["level"])
        else:
            sos = level_to_direction(max(lr1["level"], level_run[index - 1]["level"]))
        if index == len(level_run) - 1:
            eos = level_to_direction(lr1["level"])
        else:
            if lr1["text"][-1].original_bidi_class in ("LRI", "RLI", "FSI"):
                # X10 - last char is an isolator without matching PDI - set EOS to embedding level
                eos = level_to_direction(lr1["level"])
            else:
                eos = level_to_direction(
                    max(lr1["level"], level_run[index + 1]["level"])
                )
        lr1["sos"] = sos
        lr1["eos"] = eos

    # combine levels runs to create isolate runs
    isolate_runs: list[IsolatingRun] = []
    for index, lr2 in enumerate(level_run):
        if lr2["complete"]:
            continue
        sos = lr2["sos"]
        eos = lr2["eos"]
        ir_chars = lr2["text"]
        lr2["complete"] = True
        if lr2["text"][-1].original_bidi_class in ("LRI", "RLI", "FSI"):
            for nlr in level_run[index + 1 :]:
                if (
                    nlr["level"] == lr2["level"]
                    and nlr["text"][0].original_bidi_class == "PDI"
                ):
                    lr2["text"] += nlr["text"]
                    nlr["complete"] = True
                    eos = nlr["eos"]
                    if nlr["text"][-1].original_bidi_class not in ("LRI", "RLI", "FSI"):
                        break
        isolate_runs.append(IsolatingRun(characters=ir_chars, sos=sos, eos=eos))

    return isolate_runs


class BidiParagraph:
    __slots__ = (
        "text",
        "base_direction",
        "debug",
        "base_embedding_level",
        "characters",
    )

    def __init__(
        self,
        text: str,
        base_direction: Optional[TextDirection] = None,
        debug: bool = False,
    ) -> None:
        self.text = text
        self.base_direction = (
            auto_detect_base_direction(self.text, debug)
            if not base_direction
            else base_direction
        )
        self.debug = debug
        self.base_embedding_level = (
            0 if self.base_direction == TextDirection.LTR else 1
        )  # base level
        self.characters: list[BidiCharacter] = []
        self.get_bidi_characters()

    def get_characters(self) -> list[BidiCharacter]:
        return self.characters

    def get_characters_with_embedding_level(self) -> list[BidiCharacter]:
        # Calculate embedding level for each character after breaking isolating runs.
        # Only used on conformance testing
        self.reorder_resolved_levels()
        return self.characters

    def get_reordered_characters(self) -> tuple[BidiCharacter, ...]:
        return self.reorder_resolved_levels()

    def get_all(self) -> tuple[list[BidiCharacter], tuple[BidiCharacter, ...]]:
        return self.characters, self.reorder_resolved_levels()

    def get_reordered_string(self) -> str:
        "Used for conformance validation"
        return "".join(c.character for c in self.reorder_resolved_levels())

    def get_bidi_fragments(self) -> tuple[tuple[str, TextDirection], ...]:
        return self.split_bidi_fragments()

    def get_bidi_characters(self) -> None:
        # Explicit levels and directions. Rule X1

        stack: deque[DirectionalStatus] = deque()
        current_status = DirectionalStatus(
            embedding_level=self.base_embedding_level,
            directional_override_status="N",
            directional_isolate_status=False,
        )
        stack.append(replace(current_status))
        overflow_isolate_count = 0
        overflow_embedding_count = 0
        valid_isolate_count = 0
        results: list[BidiCharacter] = []

        # Explicit embeddings. Process each character individually applying rules X2 through X8
        for index, char in enumerate(self.text):
            bidi_char = BidiCharacter(
                index, char, current_status.embedding_level, self.debug
            )
            new_bidi_class = None

            if bidi_char.bidi_class == "FSI":
                bidi_char.bidi_class = (
                    "LRI"
                    if auto_detect_base_direction(
                        self.text[index + 1 :], stop_at_pdi=True, debug=self.debug
                    )
                    == TextDirection.LTR
                    else "RLI"
                )

            if bidi_char.bidi_class in ("RLE", "LRE", "RLO", "LRO", "RLI", "LRI"):
                # X2 - X5: calculate explicit embeddings and explicit overrides
                if bidi_char.bidi_class[0] == "R":
                    new_embedding_level = (
                        current_status.embedding_level + 1
                    ) | 1  # least greater odd
                else:
                    new_embedding_level = (
                        current_status.embedding_level + 2
                    ) & ~1  # least greater even
                if (
                    bidi_char.bidi_class[2] == "I"
                    and current_status.directional_override_status != "N"
                ):
                    new_bidi_class = current_status.directional_override_status
                if (
                    new_embedding_level <= MAX_DEPTH
                    and overflow_isolate_count == 0
                    and overflow_embedding_count == 0
                ):
                    current_status.embedding_level = new_embedding_level
                    current_status.directional_override_status = (
                        bidi_char.bidi_class[0]
                        if bidi_char.bidi_class[2] == "O"
                        else "N"
                    )
                    if bidi_char.bidi_class[2] == "I":
                        valid_isolate_count += 1
                        current_status.directional_isolate_status = True
                    else:
                        current_status.directional_isolate_status = False
                    stack.append(replace(current_status))
                else:
                    if bidi_char.bidi_class[2] == "I":
                        overflow_isolate_count += 1
                    else:
                        if overflow_isolate_count == 0:
                            overflow_embedding_count += 1

            if bidi_char.bidi_class not in (
                "B",
                "BN",
                "RLE",
                "LRE",
                "RLO",
                "LRO",
                "PDF",
                "FSI",
                "PDI",
            ):  # X6
                if current_status.directional_override_status != "N":
                    new_bidi_class = current_status.directional_override_status

            if bidi_char.bidi_class == "PDI":  # X6a
                if overflow_isolate_count > 0:
                    overflow_isolate_count -= 1
                elif valid_isolate_count > 0:
                    overflow_embedding_count = 0
                    while True:
                        if not stack[-1].directional_isolate_status:
                            stack.pop()
                            continue
                        break
                    stack.pop()
                    current_status = replace(stack[-1])
                    valid_isolate_count -= 1
                assert isinstance(current_status, DirectionalStatus)
                bidi_char.embedding_level = current_status.embedding_level
                if current_status.directional_override_status != "N":
                    new_bidi_class = current_status.directional_override_status

            if bidi_char.bidi_class == "PDF":  # X7
                if overflow_isolate_count == 0:
                    if overflow_embedding_count > 0:
                        overflow_embedding_count -= 1
                    else:
                        if (
                            not current_status.directional_isolate_status
                            and len(stack) > 1
                        ):
                            stack.pop()
                            current_status = replace(stack[-1])

            if new_bidi_class:
                bidi_char.bidi_class = new_bidi_class
            if bidi_char.bidi_class not in (
                "RLE",
                "LRE",
                "RLO",
                "LRO",
                "PDF",
                "BN",
            ):  # X9
                if bidi_char.bidi_class == "B":
                    bidi_char.embedding_level = self.base_embedding_level
                elif bidi_char.original_bidi_class not in ("LRI", "RLI", "FSI"):
                    bidi_char.embedding_level = current_status.embedding_level
                results.append(bidi_char)

        if not results:
            self.characters = []
            return
        self.characters = results
        calculate_isolate_runs(results)

    def split_bidi_fragments(self) -> tuple[tuple[str, TextDirection], ...]:
        bidi_fragments: list[tuple[str, TextDirection]] = []
        if len(self.characters) == 0:
            return ()
        current_fragment = ""
        current_direction = ""
        for c in self.characters:
            if c.get_direction_from_level() != current_direction:
                if current_fragment:
                    bidi_fragments.append(
                        (
                            current_fragment,
                            (
                                TextDirection.RTL
                                if current_direction == "R"
                                else TextDirection.LTR
                            ),
                        )
                    )
                current_fragment = ""
                current_direction = c.get_direction_from_level()
            current_fragment += c.character
        if current_fragment:
            bidi_fragments.append(
                (
                    current_fragment,
                    (
                        TextDirection.RTL
                        if current_direction == "R"
                        else TextDirection.LTR
                    ),
                )
            )
        return tuple(bidi_fragments)

    def reorder_resolved_levels(self) -> tuple[BidiCharacter, ...]:
        before_separator = True
        end_of_line = True
        max_level = 0
        min_odd_level = 999
        for bidi_char in reversed(self.characters):
            # Rule L1. Reset the embedding level of segment separators, paragraph separators,
            # and any adjacent whitespace.
            if bidi_char.original_bidi_class in ("S", "B"):
                bidi_char.embedding_level = self.base_embedding_level
                before_separator = True
            elif bidi_char.original_bidi_class in (
                "BN",
                "WS",
                "FSI",
                "LRI",
                "RLI",
                "PDI",
            ):
                if before_separator or end_of_line:
                    bidi_char.embedding_level = self.base_embedding_level
            else:
                before_separator = False
                end_of_line = False

            if bidi_char.embedding_level > max_level:
                max_level = bidi_char.embedding_level
            if (
                bidi_char.embedding_level % 2 != 0
                and bidi_char.embedding_level < min_odd_level
            ):
                min_odd_level = bidi_char.embedding_level

        # Rule L2. From the highest level found in the text to the lowest odd level on each line,
        # reverse any contiguous sequence of characters that are at that level or higher.
        reordered_paragraph = self.characters.copy()
        for level in range(max_level, min_odd_level - 1, -1):
            temp_results: list[BidiCharacter] = []
            rev: list[BidiCharacter] = []
            for bidi_char in reordered_paragraph:
                if bidi_char.embedding_level >= level:
                    rev.append(bidi_char)
                else:
                    if rev:
                        rev.reverse()
                        temp_results += rev
                        rev = []
                    temp_results.append(bidi_char)
            if rev:
                rev.reverse()
                temp_results += rev
            reordered_paragraph = temp_results
        return tuple(reordered_paragraph)



import pytest

from os import devnull
from pathlib import Path

from fpdf import FPDF
from fpdf.enums import XPos, YPos
from fpdf.errors import FPDFException
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_fallback_font(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(family="Roboto", style="B", fname=HERE / "Roboto-Bold.ttf")
    pdf.add_font(family="DejaVuSans", fname=HERE / "DejaVuSans.ttf")
    pdf.set_font("Roboto", size=15)
    pdf.write(text="WITHOUT FALLBACK FONT:\n")
    write_strings(pdf)

    pdf.ln(10)
    pdf.write(text="WITH FALLBACK FONT:\n")
    pdf.set_fallback_fonts(["DejaVuSans"])
    write_strings(pdf)

    assert_pdf_equal(
        pdf,
        HERE / "fallback_font.pdf",
        tmp_path,
    )


def write_strings(pdf):
    pdf.write(text="write() 😄 😁 😆 😅 ✌")
    pdf.ln()
    pdf.cell(
        text="cell() without markdown 😄 😁**bold** 😆 😅 ✌",
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.cell(
        text="cell() with markdown 😄 😁**bold** 😆 😅 ✌",
        markdown=True,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.multi_cell(
        text="multi_cell() 😄 😁 😆 😅 ✌",
        w=50,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )


def test_fallback_font_no_warning(caplog):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(family="Roboto", style="B", fname=HERE / "Roboto-Bold.ttf")
    pdf.add_font(family="DejaVuSans", fname=HERE / "DejaVuSans.ttf")
    pdf.set_font("Roboto", size=15)
    pdf.set_fallback_fonts(["DejaVuSans"])
    write_strings(pdf)
    pdf.output()
    assert len(caplog.text) == 0


def test_fallback_font_ignore_style(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(family="Roboto", style="B", fname=HERE / "Roboto-Bold.ttf")
    pdf.add_font(family="DejaVuSans", fname=HERE / "DejaVuSans.ttf")
    pdf.set_font("Roboto", size=20)

    pdf.set_fallback_fonts(["DejaVuSans"], exact_match=True)
    pdf.cell(
        text="cell() with markdown + exact_match=True: **[😄 😁]** 😆 😅 ✌",
        markdown=True,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.ln()

    pdf.set_fallback_fonts(["DejaVuSans"], exact_match=False)
    pdf.cell(
        text="cell() with markdown + exact_match=False: **[😄 😁]** 😆 😅 ✌",
        markdown=True,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.ln()

    pdf.add_font(family="DejaVuSans", style="B", fname=HERE / "DejaVuSans-Bold.ttf")
    pdf.cell(
        text="cell() with markdown + matching font style: **[😄 😁]** 😆 😅 ✌",
        markdown=True,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.ln()

    pdf.add_font(family="Roboto", style="I", fname=HERE / "Roboto-Italic.ttf")
    pdf.add_font(family="Roboto", style="BI", fname=HERE / "Roboto-BoldItalic.TTF")
    pdf.add_font(family="DejaVuSans", style="I", fname=HERE / "DejaVuSans-Oblique.ttf")
    pdf.add_font(
        family="DejaVuSans", style="IB", fname=HERE / "DejaVuSans-BoldOblique.ttf"
    )
    pdf.cell(
        text="cell() with markdown + bold-italics: __{**[😄 😁]** 😆 😅}__ ✌",
        markdown=True,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.ln()

    assert_pdf_equal(
        pdf,
        HERE / "fallback_font_ignore_style.pdf",
        tmp_path,
    )


def test_fallback_font_with_overridden_get_fallback_font(tmp_path):
    class PDF(FPDF):
        def get_fallback_font(self, char, style=""):
            fonts_with_char = [
                font_id
                for font_id in self._fallback_font_ids
                if ord(char) in self.fonts[font_id].cmap
            ]
            if not fonts_with_char:
                return None
            if len(fonts_with_char) == 1:
                return fonts_with_char[0]
            if self.font_family == "quicksand" and "dejavusans" in fonts_with_char:
                return "dejavusans"
            if self.font_family == "roboto" and "twitteremoji" in fonts_with_char:
                return "twitteremoji"
            raise NotImplementedError

    pdf = PDF()
    pdf.add_page()
    pdf.add_font(family="Quicksand", fname=HERE / "Quicksand-Regular.otf")
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(fname=HERE / "DejaVuSans.ttf")
    pdf.add_font(fname=HERE / "TwitterEmoji.ttf")
    pdf.add_font(fname=HERE / "Waree.ttf")
    text = "Hello world / สวัสดีชาวโลก ทดสอบฟอนต์, 😄 😁 😆 😅 ✌"
    pdf.set_fallback_fonts(["DejaVuSans", "TwitterEmoji", "Waree"])
    pdf.set_font("Quicksand", size=20)
    pdf.cell(
        text=text,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    pdf.ln()
    pdf.set_font("Roboto", size=20)
    pdf.cell(
        text=text,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
    )
    assert_pdf_equal(
        pdf,
        HERE / "fallback_font_with_overridden_get_fallback_font.pdf",
        tmp_path,
    )


def test_invalid_fallback_font():
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(family="Waree", fname=HERE / "Waree.ttf")
    pdf.set_font("Roboto", size=15)
    with pytest.raises(FPDFException) as error:
        pdf.set_fallback_fonts(["Waree", "Invalid"])
    assert (
        str(error.value)
        == "Undefined fallback font: Invalid - Use FPDF.add_font() beforehand"
    )


def test_glyph_not_on_any_font(caplog):
    """
    Similar to fonts\\test_add_font\\test_font_missing_glyphs,
    but resulting is less missing glyphs because the fallback font provided some of them
    """
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Roboto", fname=HERE / "Roboto-Regular.ttf")
    pdf.add_font(family="DejaVuSans", fname=HERE / "DejaVuSans.ttf")
    pdf.set_font("Roboto")
    pdf.set_fallback_fonts(["DejaVuSans"])
    pdf.cell(text="Test 𝕥𝕖𝕤𝕥 🆃🅴🆂🆃 😲")
    pdf.output(devnull)
    assert (
        "Roboto is missing the following glyphs: "
        "'🆃' (\\U0001f183), '🅴' (\\U0001f174), '🆂' (\\U0001f182)" in caplog.text
    )


def test_fallback_font_with_non_alpha_on_fontkey(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(family="Quicksand", fname=HERE / "Quicksand-Regular.otf")
    pdf.add_font(family="DejaVu-Sans", fname=HERE / "DejaVuSans.ttf")
    pdf.add_font(family="Twitter Emoji", fname=HERE / "TwitterEmoji.ttf")
    pdf.add_font(family="Waree", fname=HERE / "Waree.ttf")
    text = "Hello world / สวัสดีชาวโลก ทดสอบฟอนต์, 😄 😁 😆 😅 ✌"
    pdf.set_fallback_fonts(["DejaVu-Sans", "Twitter Emoji", "Waree"])
    pdf.set_font("Quicksand", size=20)
    pdf.cell(
        text=text,
        new_x=XPos.LMARGIN,
        new_y=YPos.NEXT,
        markdown=True,
    )
    pdf.ln()
    assert_pdf_equal(
        pdf,
        HERE / "fallback_font_with_non_alpha_on_fontkey.pdf",
        tmp_path,
    )



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent


def test_skew(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x, y = 60, 60
    img_filepath = HERE / "image/png_images/66ac49ef3f48ac9482049e1ab57a53e9.png"
    with pdf.skew(ax=45, x=x, y=y):
        pdf.image(img_filepath, x=x, y=y)
    with pdf.skew(ax=45, ay=30, x=x + 20, y=y - 20):
        pdf.image(img_filepath, x=x + 20, y=y - 20)
    pdf.set_line_width(2)
    pdf.set_draw_color(240)
    pdf.set_fill_color(r=230, g=30, b=180)
    with pdf.skew(ax=-45, ay=0, x=100, y=170):
        pdf.circle(x=105, y=175, radius=5, style="FD")
    pdf.image(img_filepath, x=150, y=150)
    assert_pdf_equal(pdf, HERE / "skew.pdf", tmp_path)


def test_skew_text(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=12)
    with doc.skew(0, 20, 20, 20):
        doc.text(20, 20, text="text skewed on the y-axis")
    with doc.skew(0, -20, 20, 60):
        doc.text(20, 60, text="text skewed on the y-axis (negative)")
        doc.text(20, 100, text="text skewed on the y-axis (negative) - line 2")
    with doc.skew(20, 0, 20, 140):
        doc.text(20, 140, text="text skewed on the x-axis")
    with doc.skew(-20, 0, 20, 180):
        doc.text(20, 180, text="text skewed on the x-axis (negative)")
    with doc.skew(89, 0, 20, 220):
        doc.text(20, 220, text="some extreme skewing")
    assert_pdf_equal(doc, HERE / "skew_text.pdf", tmp_path)


def test_cell_skew_text(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=12)
    doc.ln(40)
    with doc.skew(0, 20):
        doc.cell(text="text skewed on the y-axis")
        doc.ln(40)
    with doc.skew(0, -20):
        doc.cell(text="text skewed on the y-axis (negative)")
        doc.ln(40)
    with doc.skew(20, 0):
        doc.cell(text="text skewed on the x-axis")
        doc.ln(40)
    with doc.skew(-20, 0):
        doc.cell(text="text skewed on the x-axis (negative)")
        doc.ln(40)
    with doc.skew(89, 0):
        doc.cell(text="some extreme skewing")
        doc.ln(40)
    assert_pdf_equal(doc, HERE / "cell_skew_text.pdf", tmp_path)


def test_multi_cell_skew_text(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    # built-in font
    pdf.set_font("Helvetica", size=8)
    pdf.set_fill_color(255, 255, 0)
    with pdf.skew(20, 0):
        pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], fill=True)
        pdf.ln(60)
    with pdf.skew(0, 20):
        pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], fill=True)
        pdf.ln(60)
    with pdf.skew(20, 20):
        pdf.multi_cell(w=150, text=LOREM_IPSUM[:200], fill=True)
    assert_pdf_equal(pdf, HERE / "multi_cell_skew_text.pdf", tmp_path)



from fpdf import FPDF


class PDF(FPDF):
    def header(self):
        # Setting font: helvetica bold 15
        self.set_font("helvetica", style="B", size=15)
        # Calculating width of title and setting cursor position:
        width = self.get_string_width(self.title) + 6
        self.set_x((210 - width) / 2)
        # Setting colors for frame, background and text:
        self.set_draw_color(0, 80, 180)
        self.set_fill_color(230, 230, 0)
        self.set_text_color(220, 50, 50)
        # Setting thickness of the frame (1 mm)
        self.set_line_width(1)
        # Printing title:
        self.cell(
            width,
            9,
            self.title,
            border=1,
            new_x="LMARGIN",
            new_y="NEXT",
            align="C",
            fill=True,
        )
        # Performing a line break:
        self.ln(10)

    def footer(self):
        # Setting position at 1.5 cm from bottom:
        self.set_y(-15)
        # Setting font: helvetica italic 8
        self.set_font("helvetica", style="I", size=8)
        # Setting text color to gray:
        self.set_text_color(128)
        # Printing page number
        self.cell(0, 10, f"Page {self.page_no()}", align="C")

    def chapter_title(self, num, label):
        # Setting font: helvetica 12
        self.set_font("helvetica", size=12)
        # Setting background color
        self.set_fill_color(200, 220, 255)
        # Printing chapter name:
        self.cell(
            0,
            6,
            f"Chapter {num} : {label}",
            new_x="LMARGIN",
            new_y="NEXT",
            align="L",
            fill=True,
        )
        # Performing a line break:
        self.ln(4)

    def chapter_body(self, filepath):
        # Reading text file:
        with open(filepath, "rb") as fh:
            txt = fh.read().decode("latin-1")
        # Setting font: Times 12
        self.set_font("Times", size=12)
        # Printing justified text:
        self.multi_cell(0, 5, txt)
        # Performing a line break:
        self.ln()
        # Final mention in italics:
        self.set_font(style="I")
        self.cell(0, 5, "(end of excerpt)")

    def print_chapter(self, num, title, filepath):
        self.add_page()
        self.chapter_title(num, title)
        self.chapter_body(filepath)


pdf = PDF()
pdf.set_title("20000 Leagues Under the Seas")
pdf.set_author("Jules Verne")
pdf.print_chapter(1, "A RUNAWAY REEF", "20k_c1.txt")
pdf.print_chapter(2, "THE PROS AND CONS", "20k_c1.txt")
pdf.output("tuto3.pdf")



"""
Module dedicated to document signature generation.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/Signing.html>
"""

# pyright: reportUnknownArgumentType=false, reportUnknownMemberType=false, reportUnknownVariableType=false

import hashlib
from datetime import timezone
from typing import TYPE_CHECKING, Any, Optional
from unittest.mock import patch

from .syntax import Name, PDFDate, build_obj_dict, create_dictionary_string as pdf_dict
from .util import buffer_subst

if TYPE_CHECKING:
    from datetime import datetime

    from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
    from cryptography.x509 import Certificate
    from endesive import signer

    from .encryption import StandardSecurityHandler


class Signature:
    def __init__(
        self,
        contact_info: Optional[str] = None,
        location: Optional[str] = None,
        m: Optional[PDFDate] = None,
        reason: Optional[str] = None,
    ) -> None:
        self.type = Name("Sig")
        self.filter = Name("Adobe.PPKLite")
        self.sub_filter = Name("adbe.pkcs7.detached")
        self.contact_info = contact_info
        "Information provided by the signer to enable a recipient to contact the signer to verify the signature"
        self.location = location
        "The CPU host name or physical location of the signing"
        self.m = m
        "The time of signing"
        self.reason = reason
        "The reason for the signing"
        self.byte_range = _SIGNATURE_BYTERANGE_PLACEHOLDER
        self.contents = "<" + _SIGNATURE_CONTENTS_PLACEHOLDER + ">"

    def serialize(
        self,
        _security_handler: Optional["StandardSecurityHandler"] = None,
        _obj_id: Optional[int] = None,
    ) -> str:
        obj_dict = build_obj_dict(
            {key: getattr(self, key) for key in dir(self)},
            _security_handler=_security_handler,
            _obj_id=_obj_id,
        )
        return pdf_dict(obj_dict)


def sign_content(
    signer: "signer",  # pyright: ignore[reportGeneralTypeIssues]
    buffer: bytearray,
    key: Optional["PrivateKeyTypes"],
    cert: "Certificate",
    extra_certs: list["Certificate"],
    hashalgo: str,
    sign_time: "datetime",
) -> bytearray:
    """
    Perform PDF signing based on the content of the buffer, performing substitutions on it.
    The signing operation does not alter the buffer size
    """
    # We start by substituting the ByteRange,
    # that defines which part of the document content the signature is based on.
    # This is basically ALL the content EXCEPT the signature content itself.
    sig_placeholder = _SIGNATURE_CONTENTS_PLACEHOLDER.encode("latin1")
    start_index = buffer.find(sig_placeholder)
    end_index = start_index + len(sig_placeholder)
    content_range = (0, start_index - 1, end_index + 1, len(buffer) - end_index - 1)
    # pylint: disable=consider-using-f-string
    buffer = buffer_subst(
        buffer,
        _SIGNATURE_BYTERANGE_PLACEHOLDER,
        "[%010d %010d %010d %010d]" % content_range,
    )

    # We compute the ByteRange hash, of everything before & after the placeholder:
    content_hash = hashlib.new(hashalgo)
    content_hash.update(buffer[: content_range[1]])  # before
    content_hash.update(buffer[content_range[2] :])  # after

    # This monkey-patching is needed, at the time of endesive v2.0.9,
    # to get control over signed_time, initialized by endesive.signer.sign() to be datetime.now():
    class mock_datetime:
        @staticmethod
        def now(tz: Any) -> "datetime":  # pylint: disable=unused-argument
            return sign_time.astimezone(timezone.utc)

    sign = patch("endesive.signer.datetime", mock_datetime)(signer.sign)

    contents = sign(
        datau=None,
        key=key,
        cert=cert,
        othercerts=extra_certs,
        hashalgo=hashalgo,
        attrs=True,
        signed_value=content_hash.digest(),
    )
    contents = _pkcs11_aligned(contents).encode("latin1")
    # Sanity check, otherwise we will break the xref table:
    assert len(sig_placeholder) == len(contents)
    return buffer.replace(sig_placeholder, contents, 1)


def _pkcs11_aligned(data: tuple[int, ...]) -> str:
    data_str = "".join(f"{i:02x}" for i in data)
    return data_str + "0" * (0x4000 - len(data_str))


_SIGNATURE_BYTERANGE_PLACEHOLDER = "[0000000000 0000000000 0000000000 0000000000]"
_SIGNATURE_CONTENTS_PLACEHOLDER = _pkcs11_aligned((0,))



import logging
from io import BytesIO
from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal, ensure_rss_memory_below

from PIL import Image

HERE = Path(__file__).resolve().parent
IMAGE_PATH = HERE / "png_images/6c853ed9dacd5716bc54eb59cec30889.png"
MAX_MEMORY_MB = 12  # memory usage depends on Python version


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_warn(caplog):
    pdf = FPDF()
    pdf.oversized_images = "WARN"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=50)
    assert "OVERSIZED" in caplog.text


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscale_simple(caplog, tmp_path):
    caplog.set_level(logging.DEBUG)
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=50)
    assert "OVERSIZED: Generated new low-res image" in caplog.text
    assert len(pdf.image_cache.images) == 2, pdf.image_cache.images.keys()
    in_use_img_names = _in_use_img_names(pdf)
    assert len(in_use_img_names) == 1, pdf.image_cache.images.keys()
    assert_pdf_equal(pdf, HERE / "oversized_images_downscale_simple.pdf", tmp_path)


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscale_twice(tmp_path):
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=50)
    pdf.image(IMAGE_PATH, w=50)
    assert len(pdf.image_cache.images) == 2, pdf.image_cache.images.keys()
    in_use_img_names = _in_use_img_names(pdf)
    assert len(in_use_img_names) == 1, pdf.image_cache.images.keys()
    assert_pdf_equal(pdf, HERE / "oversized_images_downscale_twice.pdf", tmp_path)


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscaled_and_highres():
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=50)
    pdf.image(IMAGE_PATH, w=pdf.epw)
    count_of_images_in_cache = len(pdf.image_cache.images)
    assert count_of_images_in_cache == 2, pdf.image_cache.images.keys()
    # Not calling assert_pdf_equal to avoid storing a large binary (1.4M) in this git repo


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_highres_and_downscaled():
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=pdf.epw)
    pdf.image(IMAGE_PATH, w=50)
    count_of_images_in_cache = len(pdf.image_cache.images)
    assert count_of_images_in_cache == 1, pdf.image_cache.images.keys()
    # Not calling assert_pdf_equal to avoid storing a large binary (1.4M) in this git repo


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscaled_with_ratio_5(tmp_path):  # issue 1409
    def solid_png(rgb):
        buf = BytesIO()
        Image.new("RGB", (1190, 1684), rgb).save(buf, "PNG")
        buf.seek(0)
        return buf

    pdf = FPDF(format="A4", unit="pt")
    pdf.oversized_images_ratio = 5
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(solid_png(((228, 150, 150))), x=0, y=0, w=85, h=120)
    pdf.image(solid_png(((0, 150, 150))), x=100, y=0, w=85, h=120)
    pdf.image(solid_png(((228, 150, 150))), x=0, y=200, w=254, h=360)
    assert_pdf_equal(
        pdf, HERE / "oversized_images_downscaled_with_ratio_5.pdf", tmp_path
    )


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscale_biggest_1st(tmp_path):
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=50)
    pdf.image(IMAGE_PATH, w=30)
    assert len(pdf.image_cache.images) == 2, pdf.image_cache.images.keys()
    in_use_img_names = _in_use_img_names(pdf)
    assert len(in_use_img_names) == 1, pdf.image_cache.images.keys()
    assert_pdf_equal(pdf, HERE / "oversized_images_downscale_biggest_1st.pdf", tmp_path)


@ensure_rss_memory_below(MAX_MEMORY_MB)
def test_oversized_images_downscale_biggest_2nd(caplog, tmp_path):
    caplog.set_level(logging.DEBUG)
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(IMAGE_PATH, w=30)
    pdf.image(IMAGE_PATH, w=50)
    assert "OVERSIZED: Updated low-res image" in caplog.text
    assert len(pdf.image_cache.images) == 2, pdf.image_cache.images.keys()
    in_use_img_names = _in_use_img_names(pdf)
    assert len(in_use_img_names) == 1, pdf.image_cache.images.keys()
    assert_pdf_equal(pdf, HERE / "oversized_images_downscale_biggest_2nd.pdf", tmp_path)


def _in_use_img_names(pdf):
    return [name for name, img in pdf.image_cache.images.items() if img["usages"]]



from fpdf import FPDF
import pandas as pd

DF = pd.DataFrame(
    {
        "First name": ["Jules", "Mary", "Carlson", "Lucas"],
        "Last name": ["Smith", "Ramos", "Banks", "Cimon"],
        "Age": [34, 45, 19, 31],
        "City": ["San Juan", "Orlando", "Los Angeles", "Saint-Mathurin-sur-Loire"],
    }
    # Convert all data inside dataframe into string type:
).applymap(str)

COLUMNS = [list(DF)]  # Get list of dataframe columns
ROWS = DF.values.tolist()  # Get list of dataframe rows
DATA = COLUMNS + ROWS  # Combine columns and rows in one list

pdf = FPDF()
pdf.add_page()
pdf.set_font("Times", size=10)
with pdf.table(
    borders_layout="MINIMAL",
    cell_fill_color=200,  # grey
    cell_fill_mode="ROWS",
    line_height=pdf.font_size * 2.5,
    text_align="CENTER",
    width=160,
) as table:
    for data_row in DATA:
        row = table.row()
        for datum in data_row:
            row.cell(datum)
pdf.output("table_from_pandas.pdf")



#!/usr/bin/env python3

# Script to convert a GIF into an animated PDF file.
# Just for lulz.
# Script used in this article: https://chezsoi.org/lucas/blog/fpdf2-4-0-and-converting-gifs-to-pdfs.html

# REQUIRE: pip install fpdf2 imageio
# USAGE: ./gif2pdf.py $gif_filepath

import sys
from fpdf import FPDF
from imageio import mimread
from PIL import Image

in_filepath = sys.argv[1]
if not in_filepath.endswith(".gif"):
    print("Input file must be a GIF", file=sys.stderr)
    sys.exit(1)

imgs = mimread(in_filepath)
size = (imgs[0].shape[1], imgs[0].shape[0])

pdf = FPDF(format=size)
pdf.set_margin(0)
duration_in_secs = 0
for img in imgs:
    pdf.add_page(duration=duration_in_secs)
    # Converting the numpy.ndarray to a PIL.Image:
    pdf.image(Image.frombytes("RGBA", size, img.tobytes()), w=pdf.epw)
    duration_in_secs = img.meta["duration"] / 1000 or 0.04

out_filepath = in_filepath.replace(".gif", ".pdf")
pdf.output(out_filepath)
print(f'You can now open "{out_filepath}" in Adobe Acrobat Reader,')
print("press CTRL+L to launch presentation mode,")
print("and then ENTER to admire the animation!")



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_image_fit_in_rect(tmp_path):
    """
    Scale the image to fill the rectangle, keeping its aspect ratio,
    and ensure it does overflow the rectangle width or height in the process.
    """
    pdf = FPDF()
    pdf.add_page()

    rect1 = 30, 30, 60, 100
    pdf.rect(*rect1)
    pdf.image(
        HERE / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png",
        *rect1,
        keep_aspect_ratio=True,
    )

    rect2 = 100, 30, 60, 100
    pdf.rect(*rect2)
    pdf.image(
        HERE / "png_images/51a4d21670dc8dfa8ffc9e54afd62f5f.png",
        *rect2,
        keep_aspect_ratio=True,
    )

    assert_pdf_equal(pdf, HERE / "image_fit_in_rect.pdf", tmp_path)


def test_image_fit_in_rect_with_c_align_center(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.image(
        HERE / "png_images/6c853ed9dacd5716bc54eb59cec30889.png",
        w=72,
        h=102,
        x="C",
        keep_aspect_ratio=True,
    )
    assert_pdf_equal(pdf, HERE / "image_fit_in_rect_with_c_align_center.pdf", tmp_path)


def test_image_fit_in_rect_with_c_align_center_with_downscale(tmp_path):
    pdf = FPDF()
    pdf.oversized_images = "DOWNSCALE"
    pdf.add_page()
    pdf.image(
        HERE / "png_images/6c853ed9dacd5716bc54eb59cec30889.png",
        w=72,
        h=102,
        x="C",
        keep_aspect_ratio=True,
    )
    reference_pdf_path = (
        HERE / "image_fit_in_rect_with_c_align_center_with_downscale.pdf"
    )
    assert_pdf_equal(pdf, reference_pdf_path, tmp_path)
    pdf_with_full_img_path = HERE / "image_fit_in_rect_with_c_align_center.pdf"
    assert reference_pdf_path.stat().st_size < pdf_with_full_img_path.stat().st_size



#!/usr/bin/env python3

# Invoke veraPDF CLI & parse its output
# Purpose of this script:
# * abort the validation pipeline with a non-zero error code if any check fails on a PDF sample
# * aggregate all checks performed in a concise summary
# * parallelize the execution of this analysis on all PDF files
# * allow to ignore some errors considered harmless, listed in verapdf-ignore.json

# USAGE: ./verapdf.py [$pdf_filepath|--process-all-test-pdf-files|--print-aggregated-report]

import sys
from subprocess import run, PIPE, STDOUT

from scripts.checker_commons import main

CHECKS_DETAILS_URL = "https://docs.verapdf.org/validation/"
BAT_EXT = ".bat" if sys.platform in ("cygwin", "win32") else ""
FAIL_CODES = ("ERROR", "FAIL", "GRAVE", "SEVERE")
FAIL_IGNORED_PREFIXES = (  # We do not expect all PDF files to be PDF/A compliant
    "1b",
    "PDF/A Validation",
)


def analyze_pdf_file(pdf_filepath):
    try:
        command = [
            "verapdf/verapdf" + BAT_EXT,
            "--format",
            "text",
            "-v",
            pdf_filepath,
        ]
        # print(" ".join(command))
        output = run(command, check=False, stdout=PIPE, stderr=STDOUT).stdout.decode()
        # print(output)
        return pdf_filepath, parse_output(output)
    # pylint: disable=broad-exception-caught
    except BaseException as error:
        return pdf_filepath, error


def parse_output(output):
    "Parse VeraPDF CLI output into a dict."
    errors = []
    failure = ""
    warning = ""
    line_iterator = iter(output.splitlines())
    for line in line_iterator:
        if line.startswith("PASS ") or "M org.verapdf." in line:
            continue  # 1st line of every message logged by VeraPDF, containing the current time
        if line.startswith("  FAIL "):
            errors.append(line[len("  FAIL ") :])
        elif any(line.startswith(fail_code) for fail_code in FAIL_CODES):
            if failure:
                failure += " + "
            _fail_code, _filepath, error = line.split(" ", 2)
            if not any(error.startswith(prefix) for prefix in FAIL_IGNORED_PREFIXES):
                failure += line
        elif line.startswith("WARNING: "):
            if warning:
                warning += " + "
            warning += line[len("WARNING: ") :] + " - " + next(line_iterator)
            # pylint: disable=redefined-loop-name
            for line in line_iterator:
                if not line:
                    break  # WARNING stacktraces end with an empty line
                if not line.startswith("\t"):  # ignoring stacktrace
                    warning += " - " + line
        else:
            raise RuntimeError(f"Unexpected line format: {line}")
    return {"failure": failure, "warning": warning, "errors": errors}


if __name__ == "__main__":
    main("verapdf", analyze_pdf_file, sys.argv, CHECKS_DETAILS_URL)



#!/usr/bin/env python3

# Invoke Datalogics PDF Checker & parse its output
# Purpose of this script:
# * abort the validation pipeline with a non-zero error code if any check fails on a PDF sample
# * aggregate all checks performed in a concise summary
# * parallelize the execution of this analysis on all PDF files
# * allow to ignore some errors considered harmless, listed in pdfchecker-ignore.json

# Note: among the 3 checkers we use for fpdf2, PDF Checker is the only one that report errors
#       for unbalanced q/Q contexts in content streams, even if it does not provide a clear message.

# USAGE: ./pdfchecker.py [$pdf_filepath|--process-all-test-pdf-files|--print-aggregated-report]

import sys
from subprocess import check_output

from scripts.checker_commons import main

CHECKS_DETAILS_URL = "https://dev.datalogics.com/pdf-checker/the-json-profile-file/description-of-json-profile-parameters/"
UNPROCESSABLE_PDF_ERROR_LINE = "Unable to process document due to PDF Error"
CHECKER_SUMMARY_END_LINE = "<<=CHECKER_SUMMARY_END=>>"


def analyze_pdf_file(pdf_filepath):
    try:
        command = [
            "PDF_Checker/pdfchecker",
            "--profile",
            "PDF_Checker/CheckerProfiles/everything.json",
            "--input",
            pdf_filepath,
            "--password",
            "fpdf2",
        ]
        # print(" ".join(command))
        output = check_output(command).decode()
        # print(output)
        return pdf_filepath, parse_output(command, output)
    # pylint: disable=broad-exception-caught
    except BaseException as error:
        return pdf_filepath, error


def parse_output(command, output):
    """
    Parse PDF Checker indented output into a dict-tree.
    Tree leaves are empty dicts.
    """
    lines = output.splitlines()
    version = lines[0]
    if UNPROCESSABLE_PDF_ERROR_LINE in lines:
        return {
            "failure": UNPROCESSABLE_PDF_ERROR_LINE,
            "version": version,
        }
    assert CHECKER_SUMMARY_END_LINE in lines, f"{' '.join(command)} output:\n{output}"
    lines = lines[lines.index(CHECKER_SUMMARY_END_LINE) + 2 :]
    analysis = insert_indented(lines)
    return {
        "errors": [
            error
            for section in analysis.values()
            for error in section.get("Errors:", {}).keys()
            if error != "None"
        ],
        "version": version,
    }


def insert_indented(lines, node=None, depth=0, indent=0):
    if node is None:
        node = {}
    prev_node = None
    while lines:
        line = lines[0]
        if not line:
            lines.pop(0)
            continue
        line_indent = len(line) - len(line.lstrip())
        text = line[line_indent:].rstrip()
        if line_indent >= indent and text in (
            "Color Images",
            "Grayscale Images",
            "Monochrome Images",
        ):
            if depth > 1:
                # Leaving this branch of the tree after processing a "* Images" block
                return
            # Special case handled by creating a subnode for this "* Images" block:
            lines.pop(0)
            node[text] = {}
            insert_indented(lines, node[text], depth + 1, indent)
            continue
        if line_indent == indent:
            lines.pop(0)
            prev_node = node[text] = {}
            continue
        if line_indent > indent:
            if prev_node is None:
                # Case of more than 1 level of indentation, e.g. "How To Optimize:" section
                lines.pop(0)
                node[text] = {}
                continue
            assert (
                prev_node is not None
            ), f"depth={depth} indent={indent} line_indent={line_indent}: {line}"
            insert_indented(lines, prev_node, depth + 1, indent + 4)
            continue
        return  # line_indent < indent
    return node


if __name__ == "__main__":
    main("pdfchecker", analyze_pdf_file, sys.argv, CHECKS_DETAILS_URL)



from pathlib import Path

import pytest

from fpdf import FPDF, ViewerPreferences
from fpdf.syntax import Name
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_default_viewer_preferences(tmp_path):
    pdf = FPDF()
    pdf.viewer_preferences = ViewerPreferences()
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(pdf, HERE / "default_viewer_preferences.pdf", tmp_path)


def test_custom_viewer_preferences(tmp_path):
    pdf = FPDF()
    pdf.viewer_preferences = ViewerPreferences(
        hide_toolbar=True,
        hide_menubar=True,
        hide_window_u_i=True,
        fit_window=True,
        center_window=True,
        display_doc_title=True,
        non_full_screen_page_mode="USE_OUTLINES",
        print_scaling=Name("None"),
    )
    pdf.set_font("helvetica", size=30)
    pdf.add_page()
    pdf.cell(text="page 1")
    pdf.add_page()
    pdf.cell(text="page 2")
    assert_pdf_equal(pdf, HERE / "custom_viewer_preferences.pdf", tmp_path)


@pytest.mark.parametrize(
    "non_full_screen_page_mode", ("FULL_SCREEN", "USE_ATTACHMENTS")
)
def test_invalid_viewer_preferences(non_full_screen_page_mode):
    with pytest.raises(ValueError):
        ViewerPreferences(non_full_screen_page_mode=non_full_screen_page_mode)



#!/usr/bin/env python3
from itertools import chain
import fpdf

pdf = fpdf.FPDF()
pdf.set_margin(0)
for i, n in enumerate(chain(range(0x21, 0x81), range(0xA0, 0xFF))):
    if i % 25 == 0:
        col = i // 25
        if col % 3 == 0:
            pdf.l_margin = 0
            pdf.add_page()
            pdf.set_font("helvetica", size=30)
            pdf.y = 10
            pdf.cell(w=pdf.epw, text="Symbol font", align="C")
        pdf.x = pdf.l_margin = 10 + (col % 3) * 65
        pdf.y = 30
    pdf.set_font("helvetica", size=30)
    pdf.cell(text=f"\\u00{hex(n)[2:]} = ")
    pdf.set_font("symbol", size=30)
    pdf.cell(text=chr(n), ln=1)
pdf.output("symbol.pdf")



# pylint: disable=redefined-loop-name
from pathlib import Path

import pytest

import fpdf
from fpdf.enums import Align, XPos, YPos
from fpdf.line_break import MultiLineBreak, TextLine
from test.conftest import assert_pdf_equal, assert_same_file

HERE = Path(__file__).resolve().parent


CELLDATA = (
    # txt,        align, new_x,    new_y
    ("Left Top L", "L", "LEFT", "TOP"),
    ("Left Top R", "R", "LEFT", "TOP"),
    ("Left Top C", "C", "LEFT", "TOP"),
    ("Left Top X", "X", "LEFT", "TOP"),
    ("Left Top J", "J", "LEFT", "TOP"),
    ("Right Last L", "L", "RIGHT", "LAST"),
    ("Right Last R", "R", "RIGHT", "LAST"),
    ("Right Last C", "C", "RIGHT", "LAST"),
    ("Right Last X", "X", "RIGHT", "LAST"),
    ("Right Last J", "J", "RIGHT", "LAST"),
    ("Start Next L", "L", "START", "NEXT"),
    ("Start Next R", "R", "START", "NEXT"),
    ("Start Next C", "C", "START", "NEXT"),
    ("Start Next X", "X", "START", "NEXT"),
    ("Start Next J", "J", "START", "NEXT"),
    ("End TMargin L", "L", "END", "TMARGIN"),
    ("End TMargin R", "R", "END", "TMARGIN"),
    ("End TMargin C", "C", "END", "TMARGIN"),
    ("End TMargin X", "X", "END", "TMARGIN"),
    ("End TMargin J", "J", "END", "TMARGIN"),
    ("WCont Top L", "L", "WCONT", "TOP"),
    ("WCont Top R", "R", "WCONT", "TOP"),
    ("WCont Top C", "C", "WCONT", "TOP"),
    ("WCont Top X", "X", "WCONT", "TOP"),
    ("WCont Top J", "J", "WCONT", "TOP"),
    ("Center TOP L", "L", "CENTER", "TOP"),
    ("Center TOP R", "R", "CENTER", "TOP"),
    ("Center TOP C", "C", "CENTER", "TOP"),
    ("Center TOP X", "X", "CENTER", "TOP"),
    ("Center TOP J", "J", "CENTER", "TOP"),
    ("LMargin BMargin L", "L", "LMARGIN", "BMARGIN"),
    ("LMargin BMargin R", "R", "LMARGIN", "BMARGIN"),
    ("LMargin BMargin C", "C", "LMARGIN", "BMARGIN"),
    ("LMargin BMargin X", "X", "LMARGIN", "BMARGIN"),
    ("LMargin BMargin J", "J", "LMARGIN", "BMARGIN"),
    ("RMargin Top L", "L", "RMARGIN", "TOP"),
    ("RMargin Top R", "R", "RMARGIN", "TOP"),
    ("RMargin Top C", "C", "RMARGIN", "TOP"),
    ("RMargin Top X", "X", "RMARGIN", "TOP"),
    ("RMargin Top J", "J", "RMARGIN", "TOP"),
)


def test_render_styled_newpos(tmp_path):
    """
    Verify that _render_styled_text_line() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for unadorned text.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=24)
    twidth = 100

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = Align.coerce(item[1])
        newx = XPos.coerce(item[2])
        newy = YPos.coerce(item[3])
        # pylint: disable=protected-access
        frags = doc._preload_font_styles(s, False)
        mlb = MultiLineBreak(frags, twidth, [1, 1], align=align)
        line = mlb.get_line()
        if align == Align.J:
            # We need to manually rebuild our TextLine in order to force
            # justified alignment on a single line.
            line = TextLine(
                fragments=line.fragments,
                text_width=line.text_width,
                number_of_spaces=line.number_of_spaces,
                align=Align.J,
                height=0,
                max_width=line.max_width,
                trailing_nl=False,
            )
        doc._render_styled_text_line(
            line,
            border=1,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "render_styled_newpos.pdf", tmp_path)


def test_cell_newpos(tmp_path):
    """
    Verify that cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for unadorned text.

    Note:
        cell() doesn't process align="J", and uses "L" instead.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=24)
    twidth = 100

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        if align == "J":
            continue
        newx = item[2]
        newy = item[3]
        doc.cell(
            twidth,
            text=s,
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "cell_newpos.pdf", tmp_path)


def test_cell_newpos_stretched(tmp_path):
    """
    Verify that cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for stretched text.

    Note:
        cell() doesn't process align="J", and uses "L" instead.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=20)
    doc.set_stretching(150)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        if align == "J":
            continue
        newx = item[2]
        newy = item[3]
        doc.cell(
            twidth,
            text=s,
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "cell_newpos_stretched.pdf", tmp_path)


def test_cell_newpos_charspaced(tmp_path):
    """
    Verify that cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y for text with character spacing.

    Note:
        cell() doesn't process align="J", and uses "L" instead.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=20)
    doc.set_char_spacing(10)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        if align == "J":
            continue
        newx = item[2]
        newy = item[3]
        doc.cell(
            twidth,
            text=s,
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "cell_newpos_charspaced.pdf", tmp_path)


def test_cell_newpos_combined(tmp_path):
    """
    Verify that cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y for text with stretching and character spacing.

    Note:
        cell() doesn't process align="J", and uses "L" instead.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=20)
    doc.set_stretching(130)
    doc.set_char_spacing(5)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        if align == "J":
            continue
        newx = item[2]
        newy = item[3]
        doc.cell(
            twidth,
            text=s,
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "cell_newpos_combined.pdf", tmp_path)


def test_multi_cell_newpos(tmp_path):
    """
    Verify that multi_cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for unadorned text.

    Note:
        multi_cell() doesn't use align="J" on the last line, and
        uses "L" instead. new_x is relative to the last line.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=24)
    twidth = 100

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        newx = item[2]
        newy = item[3]
        doc.multi_cell(
            twidth,
            text=s + " xxxxxxxxxxxxxxx",  # force auto break
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "multi_cell_newpos.pdf", tmp_path)


def test_multi_cell_newpos_stretched(tmp_path):
    """
    Verify that multi_cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for stretched text.

    Note:
        multi_cell() doesn't use align="J" on the last line, and
        uses "L" instead. new_x is relative to the last line.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=25)
    doc.set_stretching(150)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        newx = item[2]
        newy = item[3]
        doc.multi_cell(
            twidth,
            text=s + " xxxxxxxxxxxxxxx",  # force auto break
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "multi_cell_newpos_stretched.pdf", tmp_path)


def test_multi_cell_newpos_charspaced(tmp_path):
    """
    Verify that multi_cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for text with stretching and character spacing.
    Note:
        multi_cell() doesn't use align="J" on the last line, and
        uses "L" instead. new_x is relative to the last line.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=20)
    doc.set_char_spacing(7)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        newx = item[2]
        newy = item[3]
        doc.multi_cell(
            twidth,
            text=s + " xxxxxxxxxxxx",  # force auto break
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "multi_cell_newpos_charspaced.pdf", tmp_path)


def test_multi_cell_newpos_combined(tmp_path):
    """
    Verify that multi_cell() places the new position
    in the right places in all possible combinations of alignment,
    new_x, and new_y, for text with stretching and character spacing.
    Note:
        multi_cell() doesn't use align="J" on the last line, and
        uses "L" instead. new_x is relative to the last line.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=20)
    doc.set_stretching(130)
    doc.set_char_spacing(5)
    twidth = 120

    for i, item in enumerate(CELLDATA):
        i = i % 5
        if i == 0:
            doc.add_page()
        doc.x = 70
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        newx = item[2]
        newy = item[3]
        doc.multi_cell(
            twidth,
            text=s + " xxxxxxxxxxxx",  # force auto break
            border=1,
            align=align,
            new_x=newx,
            new_y=newy,
        )
        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "multi_cell_newpos_combined.pdf", tmp_path)


LN_CELLDATA = (
    # txt, align, ln
    ("ln=0 L", "L", 0),
    ("ln=0 R", "R", 0),
    ("ln=0 C", "C", 0),
    ("ln=0 J", "J", 0),
    ("ln=1 L", "L", 1),
    ("ln=1 R", "R", 1),
    ("ln=1 C", "C", 1),
    ("ln=1 J", "J", 1),
    ("ln=2 L", "L", 2),
    ("ln=2 R", "R", 2),
    ("ln=2 C", "C", 2),
    ("ln=2 J", "J", 2),
    ("ln=3 L", "L", 3),
    ("ln=3 R", "R", 3),
    ("ln=3 C", "C", 3),
    ("ln=3 J", "J", 3),
)


def test_cell_lnpos(tmp_path):
    """
    Verify that cell() places the new position
    in the right places in all possible combinations of alignment,
    and (deprecated) ln=#.

    Note:
        cell() doesn't process align="J", and uses "L" instead.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=24)
    twidth = 100

    for i, item in enumerate(LN_CELLDATA):
        i = i % 4
        ln = item[2]
        if ln > 2:  # not valid for cell()
            break
        if i == 0:
            doc.add_page()
        doc.x = 20
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        if align == "J":
            continue
        with pytest.warns(DeprecationWarning) as record:
            doc.cell(
                twidth,
                text=s,
                border=1,
                align=align,
                ln=ln,
            )

        assert len(record) == 1
        assert_same_file(record[0].filename, __file__)

        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "cell_ln_newpos.pdf", tmp_path)


def test_multi_cell_ln_newpos(tmp_path):
    """
    Verify that multi_cell() places the new position
    in the right places in all possible combinations of alignment,
    and (deprecated) ln=#.

    Note:
        multi_cell() doesn't use align="J" on the last line, and
        uses "L" instead. new_x is relative to the last line.
    """
    doc = fpdf.FPDF()
    doc.set_font("helvetica", style="U", size=24)
    twidth = 100

    for i, item in enumerate(LN_CELLDATA):
        i = i % 4
        if i == 0:
            doc.add_page()
        doc.x = 20
        doc.y = 20 + (i * 20)
        s = item[0]
        align = item[1]
        ln = item[2]
        with pytest.warns(DeprecationWarning) as record:
            doc.multi_cell(
                twidth,
                text=s + " xxxxxxxxxxxxxxxxxxxx",  # force auto break
                border=1,
                align=align,
                ln=ln,
            )
        assert len(record) == 1
        assert_same_file(record[0].filename, __file__)

        # mark the new position in the file with crosshairs for verification
        with doc.rotation(i * -15, doc.x, doc.y):
            doc.circle(doc.x, doc.y, radius=3)
            doc.line(doc.x - 3, doc.y, doc.x + 3, doc.y)
            doc.line(doc.x, doc.y - 3, doc.x, doc.y + 3)

    assert_pdf_equal(doc, HERE / "multi_cell_ln_newpos.pdf", tmp_path)



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_code39(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.code39("*fpdf2*", x=30, y=50, w=4, h=20)
    pdf.set_font("courier", "B", size=36)
    pdf.text(x=70, y=80, text="*fpdf2*")
    assert_pdf_equal(pdf, HERE / "barcodes_code39.pdf", tmp_path)


def test_interleaved2of5(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.interleaved2of5("1337", x=65, y=50, w=4, h=20)
    pdf.set_font("courier", "B", size=36)
    pdf.text(x=80, y=80, text="1337")
    assert_pdf_equal(pdf, HERE / "barcodes_interleaved2of5.pdf", tmp_path)



"""Test for round corners rendering - including thick borders and fill artifacts"""

from pathlib import Path
import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_round_corners_thick_border(tmp_path):
    """Test that thick borders work correctly with rounded corners."""
    pdf = fpdf.FPDF()
    pdf.add_page()

    # Thick border test - should have smooth, consistent corners
    pdf.set_draw_color(0, 0, 0)
    pdf.set_fill_color(255, 255, 255)
    pdf.set_line_width(6)
    pdf.rect(50, 40, 120, 70, style="DF", round_corners=True, corner_radius=12)

    assert_pdf_equal(pdf, HERE / "round_corners_thick_border.pdf", tmp_path)


def test_round_corners_fill_no_artifacts(tmp_path):
    """Test that filled rounded corners don't show line artifacts."""
    pdf = fpdf.FPDF()
    pdf.add_page()

    # Fill with background color - should have no visible lines at corners
    pdf.set_fill_color(230, 120, 125)
    pdf.set_line_width(1)
    pdf.rect(50, 40, 120, 70, style="DF", round_corners=True, corner_radius=12)

    assert_pdf_equal(pdf, HERE / "round_corners_fill_no_artifacts.pdf", tmp_path)


def test_round_corners_combined_issue(tmp_path):
    """Test both issues together."""
    pdf = fpdf.FPDF()
    pdf.add_page()

    # Issue 1: Thick border
    pdf.set_draw_color(0, 0, 0)
    pdf.set_fill_color(255, 255, 255)
    pdf.set_line_width(6)
    pdf.rect(50, 40, 120, 70, style="DF", round_corners=True, corner_radius=12)

    # Issue 2: Fill artifacts
    pdf.set_fill_color(230, 120, 125)
    pdf.set_line_width(1)
    pdf.rect(50, 160, 120, 70, style="DF", round_corners=True, corner_radius=12)

    assert_pdf_equal(pdf, HERE / "round_corners_combined_fix.pdf", tmp_path)


def test_round_corners_draw_only(tmp_path):
    """Test rounded corners with draw only (no fill)."""
    pdf = fpdf.FPDF()
    pdf.add_page()

    # Draw only - should have smooth rounded corners
    pdf.set_draw_color(0, 0, 0)
    pdf.set_line_width(2)
    pdf.rect(50, 40, 120, 70, style="D", round_corners=True, corner_radius=12)

    assert_pdf_equal(pdf, HERE / "round_corners_draw_only.pdf", tmp_path)


def test_round_corners_fill_only(tmp_path):
    """Test rounded corners with fill only (no border)."""
    pdf = fpdf.FPDF()
    pdf.add_page()

    # Fill only - should have smooth rounded corners without border
    pdf.set_fill_color(100, 150, 200)
    pdf.rect(50, 40, 120, 70, style="F", round_corners=True, corner_radius=12)

    assert_pdf_equal(pdf, HERE / "round_corners_fill_only.pdf", tmp_path)



from fpdf.fonts import Glyph


def test_glyph_class():
    glyph = Glyph(glyph_id=32, unicode=(0,), glyph_name=".notdef", glyph_width=0)
    # pylint: disable=comparison-with-itself
    assert glyph == glyph
    assert hash(glyph) == hash(glyph)



from pathlib import Path

import pytest

from fpdf import FPDF, FPDFException, TextStyle
from test.conftest import assert_pdf_equal
from test.table.test_table import MULTILINE_TABLE_DATA

HERE = Path(__file__).resolve().parent
ROOT = HERE.parent.parent


def test_html_table_simple(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<table><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></tbody></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_simple.pdf", tmp_path)


def test_html_table_line_separators(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html(
        """<table><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></tbody></table>""",
        table_line_separators=True,
    )
    assert_pdf_equal(pdf, HERE / "html_table_line_separators.pdf", tmp_path)


def test_html_table_th_inside_tr_issue_137(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table border="1">
    <tr>
        <th width="40%">header1</th>
        <th width="60%">header2</th>
    </tr>
    <tr>
        <th>value1</th>
        <td>value2</td>
    </tr>
</table>""")
    assert_pdf_equal(pdf, HERE / "html_table_th_inside_tr_issue_137.pdf", tmp_path)


def test_html_table_with_border(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<table border="1"><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></tbody></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_border.pdf", tmp_path)


def test_html_table_with_img(caplog, tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table>
        <tr>
            <td width="50%">
                <img src="{}/test/image/png_images/affc57dfffa5ec448a0795738d456018.png" height="235" width="435"/>
            </td>
            <td width="50%">
                <img src="{}/test/image/image_types/insert_images_insert_png.png" height="162" width="154"/>
            </td>
        </tr>
    </table>""".replace("{}", str(ROOT)))
    assert_pdf_equal(pdf, HERE / "html_table_with_img.pdf", tmp_path)
    assert 'Ignoring unsupported "width" / "height" set on <img> element' in caplog.text


def test_html_table_with_img_without_explicit_dimensions(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table>
        <tr>
            <td width="50%">
                <img src="{}/test/image/png_images/affc57dfffa5ec448a0795738d456018.png"/>
            </td>
            <td width="50%">
                <img src="{}/test/image/image_types/insert_images_insert_png.png"/>
            </td>
        </tr>
    </table>""".replace("{}", str(ROOT)))
    assert_pdf_equal(
        pdf,
        HERE / "html_table_with_img_without_explicit_dimensions.pdf",
        tmp_path,
    )


def test_html_table_with_imgs_captions_and_colspan(caplog, tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table border="0">
        <tr>
            <td colspan="2" align="center"><b>Side by side centered pictures and captions</b></td>
        </tr>
        <tr>
            <td width="50%" align="center"><img src="{}/docs/fpdf2-logo.png"/></td>
            <td width="50%" align="center"><img src="{}/docs/fpdf2-logo.png"/></td>
        </tr>
        <tr>
            <td width="50%" align="center">left caption</td>
            <td width="50%" align="center">right caption</td>
        </tr>
    </table>""".replace("{}", str(ROOT)))
    assert_pdf_equal(
        pdf, HERE / "html_table_with_imgs_captions_and_colspan.pdf", tmp_path
    )
    assert (
        'Ignoring width="50%" specified on a <td> that is not in the first <tr>'
        in caplog.text
    )


def test_html_table_with_empty_cell_contents(tmp_path):  # issue 349
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    # Reference table cells positions:
    pdf.write_html("""<table><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></tbody></table>""")
    # Table with empty cells:
    pdf.write_html("""<table><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td></td><td>3</td>
    </tr><tr>
        <td></td><td>5</td><td></td>
    </tr></tbody></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_empty_cell_contents.pdf", tmp_path)


def test_html_table_with_bgcolor(tmp_path):  # issue-512
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table>
    <thead>
        <tr>
            <th width="25%">Career</th>
            <th width="75%">Quote</th>
        </tr>
    </thead>
    <tbody>
        <tr bgcolor="grey"><td>Engineer</td><td>The engineer has been, and is, a maker of history.</td></tr>
        <tr bgcolor="white"><td>Developer</td><td>Logical thinking, passion and perseverance is the paint on your palette.</td></tr>
        <tr bgcolor="grey"><td>Analyst</td><td>Seeing what other people can't see gives you great vision.</td></tr>
        <tr bgcolor="white"><td><i>None of the above</i></td><td>I'm sorry. We could not find a quote for your job.</td></tr>
    </tbody>
</table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_bgcolor.pdf", tmp_path)


def test_html_table_with_only_tds(tmp_path):  # issue-740
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<table><tr>
        <td>left</td><td>center</td><td>right</td>
    </tr><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_only_tds.pdf", tmp_path)


def test_html_table_with_multi_lines_text(tmp_path):  # issue-91
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    pdf.write_html("""<table border="1"><thead><tr>
    <th width="30%">First name</th><th width="30%">Last name</th><th width="15%">Age</th><th width="25%">City</th>
</tr></thead><tbody><tr>
    <td>Jean Abdul William</td><td>Smith</td><td>34</td><td>San Juan</td>
</tr></tbody></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_multi_lines_text.pdf", tmp_path)


def test_html_table_with_multiline_cells_and_split_over_page(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    html = "<table><thead><tr>"
    # pylint: disable=consider-using-join
    for cell_text in MULTILINE_TABLE_DATA[0]:
        html += f"\n<th>{cell_text}</th>"
    html += "\n</tr></thead><tbody>"
    for data_row in MULTILINE_TABLE_DATA[1:-1] + MULTILINE_TABLE_DATA[1:]:
        html += "\n<tr>"
        for cell_text in data_row:
            html += f"\n<td>{cell_text}</td>"
        html += "</tr>"
    html += "\n</tbody></table>"
    pdf.write_html(html)
    assert_pdf_equal(
        pdf, HERE / "html_table_with_multiline_cells_and_split_over_page.pdf", tmp_path
    )


def test_html_table_with_width_and_align(tmp_path):
    pdf = FPDF()
    pdf.set_font_size(24)
    pdf.add_page()
    pdf.write_html("""<table width=50% align=right><thead><tr>
        <th width="25%">left</th><th width="50%">center</th><th width="25%">right</th>
    </tr></thead><tbody><tr>
        <td>1</td><td>2</td><td>3</td>
    </tr><tr>
        <td>4</td><td>5</td><td>6</td>
    </tr></tbody></table>""")
    assert_pdf_equal(pdf, HERE / "html_table_with_width_and_align.pdf", tmp_path)


def test_html_table_invalid(caplog):
    pdf = FPDF()
    pdf.set_font_size(30)
    pdf.add_page()
    # OK with empty tables:
    pdf.write_html("<table></table>")
    pdf.write_html("<table><tr></tr></table>")
    pdf.write_html("<table><tr><td></td></tr></table>")
    # KO if some elements are missing:
    with pytest.raises(FPDFException) as error:
        pdf.write_html("<table><td></td></table>")
    assert str(error.value) == "Invalid HTML: <td> used outside any <tr>"
    with pytest.raises(FPDFException) as error:
        pdf.write_html("<table><th></th></table>")
    assert str(error.value) == "Invalid HTML: <th> used outside any <tr>"
    with pytest.raises(FPDFException) as error:
        pdf.write_html("<tr></tr>")
    assert str(error.value) == "Invalid HTML: <tr> used outside any <table>"
    assert caplog.text == ""


def test_html_table_with_nested_tags():  # issue 845
    pdf = FPDF()
    pdf.set_font_size(24)
    pdf.add_page()
    with pytest.raises(NotImplementedError):
        pdf.write_html("""<table><tr>
            <th>LEFT</th>
            <th>RIGHT</th>
        </tr><tr>
            <td><font size=7>This is supported</font></td>
            <td>This <font size=20>is not</font> <b>supported</b></td>
        </tr></table>""")


def test_html_table_with_font_tags_used_to_set_text_color(tmp_path):  # issue 1013
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=8)
    pdf.write_html("""<table>
    <thead>
        <tr bgcolor="#711C45">
            <th width="25%"><font color="#711C45">Mark</font></th>
        </tr>
    </thead>
    <tbody>
        <tr bgcolor="#E4E4E4">
            <td align="center"><font color="#E4E4E4">less than 2</font></td>
        </tr>
    </tbody>
    </table>""")
    assert_pdf_equal(
        pdf,
        HERE / "html_table_with_font_tags_used_to_set_text_color.pdf",
        tmp_path,
    )


def test_html_table_with_data_that_contains_entity_names(tmp_path):  # issue 1010
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=8)
    pdf.write_html("""<table>
    <thead>
        <tr>
            <th>Value</th>
        </tr>
    </thead>
    <tbody>
        <tr>
            <td>Pi &lt; 22 &divide; 7</td>
        </tr>
    </tbody>
    </table>""")
    assert_pdf_equal(
        pdf,
        HERE / "html_table_with_data_that_contains_entity_names.pdf",
        tmp_path,
    )


def test_html_table_honoring_align(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<table>
          <tr>
            <th align="right">Foo</th>
            <th align="center">Bar</th>
            <th align="left">Baz</th>
          </tr>
          <tr>
            <td align="right">Foo</td>
            <td align="center">Bar</td>
            <td align="left">Baz</td>
          </tr>
        </table>""")
    assert_pdf_equal(pdf, HERE / "html_table_honoring_align.pdf", tmp_path)


def test_html_table_with_null_text_in_span_cell(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """<table border="1">
        <tr>
            <td rowspan=2></td>
            <td>cell 2</td>
            <td>cell 3</td>
        </tr>
        <tr>
            <td>cell 4</td>
            <td>cell 5</td>
        </tr>
        <tr>
            <td colspan=2></td>
            <td>cell 7</td>
        </tr>
    </table>
        """,
        table_line_separators=True,
    )
    assert_pdf_equal(
        pdf,
        HERE / "html_table_with_null_text_in_span_cell.pdf",
        tmp_path,
    )


def test_html_table_inside_paragraph(tmp_path):  # issue 1453
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html(
        """
<h1>This is a &lt;h1&gt;</h1>
<p>This is a &lt;p&gt;</p>
Table outside paragraph:
<table>
 <tr><th>th1</th><th>th2</th></tr>
 <tr align="C"><td>td1</td><td>td2</td></tr>
</table>
<p>
  Table inside paragraph:
  <table>
   <tr><th>th1</th><th>th2</th></tr>
   <tr align="C"><td>td1</td><td>td2</td></tr>
  </table>
</p>""",
        tag_styles={
            "h1": TextStyle(font_size_pt=32),
            "p": TextStyle(font_size_pt=16),
        },
    )
    assert_pdf_equal(pdf, HERE / "html_table_inside_paragraph.pdf", tmp_path)



"""Charmap Test Case

This module contains the test case for the Charmap Test. It prints the first
999 characters of the unicode character set with a unicode ttf font, and
verifies the result against a known good result.

This test will complain that some of the values in this font file are out of
the range of the C 'short' data type (2 bytes, 0 - 65535):
  fpdf/ttfonts.py:671: UserWarning: cmap value too big/small:
and this seems to be okay.
"""

import logging
from pathlib import Path

from fontTools.ttLib import TTFont
import pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal, ensure_exec_time_below

HERE = Path(__file__).resolve().parent


@pytest.mark.parametrize(
    "font_filename",
    [
        font_file.name
        for font_file in HERE.glob("*.*tf")
        if not any(
            exclude in font_file.stem
            for exclude in ("Bold", "Italic", "NotoColorEmoji")
        )
    ],
)
@ensure_exec_time_below(seconds=10)  # TwitterEmoji.ttf is the longest to process
def test_charmap_first_999_chars(caplog, font_filename, tmp_path):
    """
    Character Map Test
    from PyFPDF version 1.7.2: github.com/reingart/pyfpdf/commit/2eab310cfd866ce24947c3a9d850ebda7c6d515d
    """
    caplog.set_level(logging.ERROR)  # hides fonttool warnings

    font_path = HERE / font_filename
    font_name = font_path.stem

    pdf = FPDF()
    pdf.add_page()
    # disable color fonts for this test due to color glyph count limit
    pdf.render_color_fonts = False
    pdf.add_font(fname=font_path)
    pdf.set_font(font_name, size=10)

    font = TTFont(font_path, lazy=True)
    cmap = font.getBestCmap()

    # Create a PDF with the first 999 characters defined in the font:
    for counter, character in enumerate(list(cmap.keys())[:1000]):
        pdf.write(8, f"{counter:03}) {character:03x} - {character:c}", print_sh=True)
        pdf.ln()

    assert_pdf_equal(pdf, HERE / f"charmap_first_999_chars-{font_name}.pdf", tmp_path)



from markdown import markdown

html = markdown(
    """
# Top title (ATX)

Subtitle (setext)
-----------------

### An even lower heading (ATX)

**Text in bold**

_Text in italics_

[This is a link](https://github.com/PyFPDF/fpdf2)

<https://py-pdf.github.io/fpdf2/>

This is an unordered list:

* an item
* another item

This is an ordered list:

1. first item
2. second item
3. third item with an unordered sublist:
    * an item
    * another item

Inline `code span`

A table:

Foo | Bar | Baz
--- | --- | ---
Foo | Bar | Baz

Definition list:

Term
: Definition

Actual HTML:

<dl>
  <dt>Term1</dt><dd>Definition1</dd>
  <dt>Term2</dt><dd>Definition2</dd>
</dl>

Some horizontal thematic breaks:

***

---

___

![Alternate description](https://py-pdf.github.io/fpdf2/fpdf2-logo.png)
""",
    extensions=["def_list", "sane_lists", "tables"],
)

from fpdf import FPDF

pdf = FPDF()
pdf.add_page()
pdf.write_html(html)
pdf.output("pdf-from-markdown-with-markdown.pdf")



from io import BytesIO
from fpdf import FPDF
import pandas as pd
import matplotlib.pyplot as plt
import io

DATA = {
    "Unemployment_Rate": [6.1, 5.8, 5.7, 5.7, 5.8, 5.6, 5.5, 5.3, 5.2, 5.2],
    "Stock_Index_Price": [1500, 1520, 1525, 1523, 1515, 1540, 1545, 1560, 1555, 1565],
}
COLUMNS = tuple(DATA.keys())

plt.figure()  # Create a new figure object
df = pd.DataFrame(DATA, columns=COLUMNS)
df.plot(x=COLUMNS[0], y=COLUMNS[1], kind="scatter")

# Converting Figure to an image:
img_buf = BytesIO()  # Create image object
plt.savefig(img_buf, dpi=200)  # Save the image

pdf = FPDF()
pdf.add_page()
pdf.image(img_buf, w=pdf.epw)  # Make the image full width
pdf.output("matplotlib_pandas.pdf")
img_buf.close()



from fpdf import FPDF
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import numpy as np
from PIL import Image

fig = Figure(figsize=(6, 2), dpi=300)
gca = fig.gca()
gca.text(0, 0.5, r"$x^n + y^n = \frac{a}{b}$", fontsize=60)
gca.axis("off")

canvas = FigureCanvas(fig)
canvas.draw()
img = Image.fromarray(np.asarray(canvas.buffer_rgba()))

pdf = FPDF()
pdf.add_page()
pdf.image(img, w=100)
pdf.output("equation_matplotlib_raster.pdf")



elements = [
    {
        "name": "multi",
        "type": "T",
        "x1": 80,
        "y1": 10,
        "x2": 170,
        "y2": 15,
        "text": "If multiline is False, the text should still not wrap even if wrapmode is specified.",
        "background": 0xEEFFFF,
        "multiline": False,
        "wrapmode": "WORD",
    },
    {
        "name": "multi",
        "type": "T",
        "x1": 80,
        "y1": 40,
        "x2": 170,
        "y2": 45,
        "text": "If multiline is True, and wrapmode is omitted, the text should wrap by word and not "
        + "cause an error due to omission of the wrapmode argument.",
        "background": 0xEEFFFF,
        "multiline": True,
    },
    {
        "name": "multi",
        "type": "T",
        "x1": 80,
        "y1": 70,
        "x2": 170,
        "y2": 75,
        "text": "If multiline is True and the wrapmode argument is provided, it should not cause a "
        + "problem even if using the default wrapmode of 'WORD'.",
        "background": 0xEEFFFF,
        "multiline": True,
        "wrapmode": "WORD",
    },
    {
        "name": "multi",
        "type": "T",
        "x1": 80,
        "y1": 100,
        "x2": 170,
        "y2": 105,
        "text": "If multiline is True and the wrapmode is 'CHAR', it should result in "
        + "wrapping based on characters instead of words, regardless of language (i.e. even though "
        + "this is designed to support scripts like Chinese and Japanese, wrapping long sentences "
        + "like this in English still demonstrates functionality.)",
        "background": 0xEEFFFF,
        "multiline": True,
        "wrapmode": "CHAR",
    },
]



from io import BytesIO
from fpdf import FPDF
from matplotlib.figure import Figure

fig = Figure(figsize=(6, 2))
gca = fig.gca()
gca.text(0, 0.5, r"$x^n + y^n = \frac{a}{b}$", fontsize=60)
gca.axis("off")

# Converting Figure to a SVG image:
img = BytesIO()
fig.savefig(img, format="svg")

pdf = FPDF()
pdf.add_page()
pdf.image(img, w=100)
pdf.output("equation_matplotlib.pdf")



from fpdf import FPDF


class PDF(FPDF):
    def header(self):
        self.set_font("helvetica", style="B", size=15)
        width = self.get_string_width(self.title) + 6
        self.set_x((210 - width) / 2)
        self.set_draw_color(0, 80, 180)
        self.set_fill_color(230, 230, 0)
        self.set_text_color(220, 50, 50)
        self.set_line_width(1)
        self.cell(
            width,
            9,
            self.title,
            border=1,
            new_x="LMARGIN",
            new_y="NEXT",
            align="C",
            fill=True,
        )
        self.ln(10)

    def footer(self):
        self.set_y(-15)
        self.set_font("helvetica", style="I", size=8)
        self.set_text_color(128)
        self.cell(0, 10, f"Page {self.page_no()}", align="C")

    def chapter_title(self, num, label):
        self.set_font("helvetica", size=12)
        self.set_fill_color(200, 220, 255)
        self.cell(
            0,
            6,
            f"Chapter {num} : {label}",
            new_x="LMARGIN",
            new_y="NEXT",
            border="L",
            fill=True,
        )
        self.ln(4)

    def chapter_body(self, fname):
        # Reading text file:
        with open(fname, "rb") as fh:
            txt = fh.read().decode("latin-1")
        with self.text_columns(
            ncols=3, gutter=5, text_align="J", line_height=1.19
        ) as cols:
            # Setting font: Times 12
            self.set_font("Times", size=12)
            cols.write(txt)
            cols.ln()
            # Final mention in italics:
            self.set_font(style="I")
            cols.write("(end of excerpt)")

    def print_chapter(self, num, title, fname):
        self.add_page()
        self.chapter_title(num, title)
        self.chapter_body(fname)


pdf = PDF()
pdf.set_title("20000 Leagues Under the Seas")
pdf.set_author("Jules Verne")
pdf.print_chapter(1, "A RUNAWAY REEF", "20k_c1.txt")
pdf.print_chapter(2, "THE PROS AND CONS", "20k_c1.txt")
pdf.output("tuto4.pdf")



"""Test that _render_styled_text_line() correctly handles sequences of
Fragments with varying font/size/style and other characteristics.
"""

# pylint: disable=protected-access
from pathlib import Path

from fpdf import FPDF
from fpdf.enums import Align, XPos, YPos
from fpdf.line_break import Fragment, MultiLineBreak
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


TEXT_1 = "Lorem ipsum Ut nostrud irure"
TEXT_2 = " reprehenderit anim nostrud dolore sed ut Excepteur dolore ut"
TEXT_3 = " sunt irure consectetur tempor eu tempor nostrud dolore sint"
TEXT_4 = " exercitation aliquip velit ullamco esse dolore mollit"
TEXT_5 = " ea sed voluptate commodo amet eiusmod incididunt"


class FxFPDF(FPDF):
    def write_fragments(self, frags, align=Align.L):
        """Replicate the part of write() that actually renders the fragments."""

        def _get_width(height):  # pylint: disable=unused-argument
            return max_width

        text_lines = []
        multi_line_break = MultiLineBreak(frags, _get_width, [1, 1], align=align)
        # first line from current x position to right margin
        max_width = self.w - self.x - self.r_margin
        text_line = multi_line_break.get_line()
        # remaining lines fill between margins
        max_width = self.w - self.l_margin - self.r_margin
        while (text_line) is not None:
            text_lines.append(text_line)
            text_line = multi_line_break.get_line()
        if text_line:
            text_lines.append(text_line)
        if not text_lines:
            return
        self.ws = 0  # currently only left aligned, so no word spacing
        for text_line_index, text_line in enumerate(text_lines):
            if text_line_index > 0:
                self.ln()
            self._render_styled_text_line(
                text_line,
                h=None,
                border=0,
                new_x=XPos.WCONT,
                new_y=YPos.TOP,
                fill=False,
                link=None,
            )


def test_varfrags_fonts(tmp_path):
    pdf = FxFPDF()
    pdf.add_page()
    pdf.line(pdf.l_margin, pdf.t_margin, pdf.l_margin, pdf.h - pdf.b_margin)
    pdf.line(
        pdf.l_margin + pdf.epw,
        pdf.t_margin,
        pdf.l_margin + pdf.epw,
        pdf.h - pdf.b_margin,
    )
    pdf.add_font("Roboto", "B", HERE / "../fonts/Roboto-Bold.ttf")
    pdf.set_font("helvetica", size=14)
    f1 = Fragment(TEXT_1, pdf._get_current_graphics_state(), pdf.k)
    pdf.set_font("courier", style="I", size=16)
    f2 = Fragment(TEXT_2, pdf._get_current_graphics_state(), pdf.k)
    f3 = Fragment(TEXT_3, f1.graphics_state, pdf.k)
    pdf.set_font("roboto", style="B", size=20)
    f4 = Fragment(TEXT_4, pdf._get_current_graphics_state(), pdf.k)
    f5 = Fragment(TEXT_5, f1.graphics_state, pdf.k)
    frags = [f1, f2, f3, f4, f5]
    pdf.write_fragments(frags)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.J)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.R)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.C)
    assert_pdf_equal(pdf, HERE / "varfrags_fonts.pdf", tmp_path)


def test_varfrags_size_bold(tmp_path):
    pdf = FxFPDF()
    pdf.add_page()
    pdf.line(pdf.l_margin, pdf.t_margin, pdf.l_margin, pdf.h - pdf.b_margin)
    pdf.line(
        pdf.l_margin + pdf.epw,
        pdf.t_margin,
        pdf.l_margin + pdf.epw,
        pdf.h - pdf.b_margin,
    )
    pdf.set_font("helvetica", size=14)
    f1 = Fragment(TEXT_1, pdf._get_current_graphics_state(), pdf.k)
    pdf.set_font("helvetica", style="B", size=30)
    f2 = Fragment(TEXT_2, pdf._get_current_graphics_state(), pdf.k)
    f3 = Fragment(TEXT_3, f1.graphics_state, pdf.k)
    f4 = Fragment(TEXT_4, f2.graphics_state, pdf.k)
    f5 = Fragment(TEXT_5, f1.graphics_state, pdf.k)
    frags = [f1, f2, f3, f4, f5]
    pdf.write_fragments(frags)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.J)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.R)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.C)
    assert_pdf_equal(pdf, HERE / "varfrags_size_bold.pdf", tmp_path)


def test_varfrags_stretch_spacing(tmp_path):
    pdf = FxFPDF()
    pdf.add_page()
    pdf.line(pdf.l_margin, pdf.t_margin, pdf.l_margin, pdf.h - pdf.b_margin)
    pdf.line(
        pdf.l_margin + pdf.epw,
        pdf.t_margin,
        pdf.l_margin + pdf.epw,
        pdf.h - pdf.b_margin,
    )
    pdf.set_font("helvetica", size=14)
    f1 = Fragment(TEXT_1, pdf._get_current_graphics_state(), pdf.k)
    pdf.set_stretching(140)
    f2 = Fragment(TEXT_2, pdf._get_current_graphics_state(), pdf.k)
    f3 = Fragment(TEXT_3, f1.graphics_state, pdf.k)
    pdf.set_char_spacing(-3)
    f4 = Fragment(TEXT_4, pdf._get_current_graphics_state(), pdf.k)
    f5 = Fragment(TEXT_5, f1.graphics_state, pdf.k)
    frags = [f1, f2, f3, f4, f5]
    pdf.write_fragments(frags)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.J)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.R)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.C)
    assert_pdf_equal(pdf, HERE / "varfrags_stretch_spacing.pdf", tmp_path)


def test_varfrags_text_mode(tmp_path):
    pdf = FxFPDF()
    pdf.add_page()
    pdf.line(pdf.l_margin, pdf.t_margin, pdf.l_margin, pdf.h - pdf.b_margin)
    pdf.line(
        pdf.l_margin + pdf.epw,
        pdf.t_margin,
        pdf.l_margin + pdf.epw,
        pdf.h - pdf.b_margin,
    )
    pdf.set_draw_color(255, 100, 0)
    pdf.line_width = 1
    pdf.set_font("helvetica", style="B", size=20)
    f1 = Fragment(TEXT_1, pdf._get_current_graphics_state(), pdf.k)
    pdf.text_mode = "STROKE"
    f2 = Fragment(TEXT_2, pdf._get_current_graphics_state(), pdf.k)
    f3 = Fragment(TEXT_3, f1.graphics_state, pdf.k)
    # For now, the most recently set color before write_fragments() wins.
    pdf.set_draw_color(0, 0, 255)
    pdf.text_mode = "FILL_STROKE"
    f4 = Fragment(TEXT_4, pdf._get_current_graphics_state(), pdf.k)
    f5 = Fragment(TEXT_5, f1.graphics_state, pdf.k)
    frags = [f1, f2, f3, f4, f5]
    pdf.write_fragments(frags)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.J)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.R)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.C)
    assert_pdf_equal(pdf, HERE / "varfrags_text_mode.pdf", tmp_path)


def test_varfrags_char_vpos(tmp_path):
    pdf = FxFPDF()
    pdf.add_page()
    pdf.line(pdf.l_margin, pdf.t_margin, pdf.l_margin, pdf.h - pdf.b_margin)
    pdf.line(
        pdf.l_margin + pdf.epw,
        pdf.t_margin,
        pdf.l_margin + pdf.epw,
        pdf.h - pdf.b_margin,
    )
    pdf.set_font("helvetica", size=14)
    f1 = Fragment(TEXT_1, pdf._get_current_graphics_state(), pdf.k)
    pdf.char_vpos = "SUB"
    f2 = Fragment(TEXT_2, pdf._get_current_graphics_state(), pdf.k)
    f3 = Fragment(TEXT_3, f1.graphics_state, pdf.k)
    pdf.char_vpos = "SUP"
    f4 = Fragment(TEXT_4, pdf._get_current_graphics_state(), pdf.k)
    f5 = Fragment(TEXT_5, f1.graphics_state, pdf.k)
    frags = [f1, f2, f3, f4, f5]
    pdf.write_fragments(frags)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.J)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.R)
    pdf.ln()
    pdf.ln()
    pdf.write_fragments(frags, align=Align.C)
    assert_pdf_equal(pdf, HERE / "varfrags_char_vpos.pdf", tmp_path)



"""
Utilities to manage deprecation errors & warnings.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.
"""

import contextlib
import inspect
import os.path
import warnings
from functools import wraps
from types import ModuleType
from typing import Any, Callable, Iterable, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")


def support_deprecated_txt_arg(fn: Callable[P, R]) -> Callable[P, R]:
    """Decorator converting `txt=` arguments into `text=` arguments"""

    @wraps(fn)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        txt_value = kwargs.pop("txt", None)
        if txt_value is not None:
            if "text" in kwargs:
                raise ValueError("Both txt= & text= arguments cannot be provided")
            kwargs["text"] = txt_value
            warnings.warn(
                'The parameter "txt" has been renamed to "text" in 2.7.6',
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
        return fn(*args, **kwargs)

    return wrapper


def deprecated_parameter(
    parameters: Iterable[tuple[str, str]],
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator removing deprecated keyword arguments from a function call."""

    deprecated_info = tuple(parameters)
    _sentinel = object()

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        @wraps(fn)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            for name, version in deprecated_info:
                if kwargs.pop(name, _sentinel) is not _sentinel:
                    warnings.warn(
                        f'"{name}" parameter is deprecated since v{version} and will be removed in a future release',
                        DeprecationWarning,
                        stacklevel=get_stack_level(),
                    )
            return fn(*args, **kwargs)

        return wrapper

    return decorator


class WarnOnDeprecatedModuleAttributes(ModuleType):
    def __call__(self) -> None:
        raise TypeError(
            "You tried to instantied the fpdf module."
            " You probably want to import the FPDF class instead:"
            " from fpdf import FPDF"
        )

    def __getattr__(self, name: str) -> Any:
        if name in ("FPDF_CACHE_DIR", "FPDF_CACHE_MODE"):
            warnings.warn(
                (
                    "fpdf.FPDF_CACHE_DIR & fpdf.FPDF_CACHE_MODE"
                    " have been deprecated in favour of"
                    " FPDF(font_cache_dir=...)"
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            return None
        return super().__getattribute__(name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("FPDF_CACHE_DIR", "FPDF_CACHE_MODE"):
            warnings.warn(
                (
                    "fpdf.FPDF_CACHE_DIR & fpdf.FPDF_CACHE_MODE"
                    " have been deprecated in favour of"
                    " FPDF(font_cache_dir=...)"
                ),
                DeprecationWarning,
                stacklevel=get_stack_level(),
            )
            return
        super().__setattr__(name, value)


def get_stack_level() -> int:
    """Get the first place in the call stack that is not inside fpdf2"""

    # pylint: disable=import-outside-toplevel
    import fpdf  # pylint: disable=cyclic-import

    pkg_dir = os.path.dirname(fpdf.__file__)
    contextlib_dir = os.path.dirname(contextlib.__file__)

    frame = inspect.currentframe()
    n = 0
    while frame is not None:
        fname = inspect.getfile(frame)
        if fname.startswith(pkg_dir) or fname.startswith(contextlib_dir):
            frame = frame.f_back
            n += 1
        else:
            break
    return n
